#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for schema-based configuration validation."""

import copy
import io
import logging
import os
import textwrap
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import fastjsonschema
from deepmerge import Merger, always_merger
from deepmerge.strategy.dict import DictStrategies
from deepmerge.strategy.list import ListStrategies
from deepmerge.strategy.set import SetStrategies
from ruamel.yaml import YAML, YAMLError
from ruamel.yaml.comments import CommentedMap as CMap
from ruamel.yaml.comments import CommentedSeq as CSeq

from spsdk import SPSDK_YML_INDENT, SPSDKError
from spsdk.utils.misc import find_file, load_configuration, value_to_int, write_file

ENABLE_DEBUG = False

logger = logging.getLogger(__name__)


class SPSDKListStrategies(ListStrategies):
    """Extended List Strategies."""

    # pylint: disable=unused-argument   # because of the base class
    @staticmethod
    def strategy_set(_config, _path, base, nxt):  # type: ignore
        """Use the set of both as a output."""
        try:
            ret = list(set(base + nxt))
            ret.sort()
        except TypeError:
            try:
                ret = base + nxt
            except TypeError:
                logger.warning(
                    "Found unhashable object in List 'set' strategy during merge."
                    " It was used 'override' method instead of 'set'."
                )
                ret = nxt
        return ret


class SPSDKMerger(Merger):
    """Modified Merger to add new list strategy 'set'."""

    PROVIDED_TYPE_STRATEGIES = {
        list: SPSDKListStrategies,
        dict: DictStrategies,
        set: SetStrategies,
    }


def _is_number(param: Any) -> bool:
    """Checks whether the input represents a number.

    :param param: Input to analyze
    :raises SPSDKError: Input doesn't represent a number
    :return: True if input represents a number
    """
    try:
        value_to_int(param)
        return True
    except SPSDKError:
        return False


def _print_validation_fail_reason(
    exc: fastjsonschema.JsonSchemaValueException,
    extra_formaters: Optional[Dict[str, Callable[[str], bool]]] = None,
) -> str:
    """Print formated and easy to read reason why the  validation failed.

    :param exc: Original exception.
    :param extra_formaters: Additional custom formatters
    :return: String explaining the reason of fail.
    """

    def process_nested_rule(
        exception: fastjsonschema.JsonSchemaValueException,
        extra_formaters: Optional[Dict[str, Callable[[str], bool]]],
    ) -> str:
        message = ""
        for rule_def_ix, rule_def in enumerate(exception.rule_definition):
            try:
                validator = fastjsonschema.compile(rule_def, formats=extra_formaters)
                validator(exception.value)
            except fastjsonschema.JsonSchemaValueException as _exc:
                message += (
                    f"\nReason of fail for {exception.rule} rule#{rule_def_ix}: "
                    f"\n {_print_validation_fail_reason(_exc , extra_formaters)}\n"
                )
        if all(rule_def.get("required") for rule_def in exception.rule_definition):
            message += f"\nYou need to define {exception.rule} of the following sets:"
            for rule_def in exc.rule_definition:
                message += f" {rule_def['required']}"
        return message

    message = str(exc)
    if exc.rule == "required":
        missing = filter(lambda x: x not in exc.value.keys(), exc.rule_definition)
        message += f"; Missing field(s): {', '.join(missing)}"
    elif exc.rule == "format":
        if exc.rule_definition == "file":
            message += f"; Non-existing file: {exc.value}"
            message += "; The file must exists even if the key is NOT used in configuration."
    elif exc.rule == "anyOf":
        message += process_nested_rule(exc, extra_formaters=extra_formaters)
    elif exc.rule == "oneOf":
        message += process_nested_rule(exc, extra_formaters=extra_formaters)
    return message


def check_config(
    config: Union[str, Dict[str, Any]],
    schemas: List[Dict[str, Any]],
    extra_formaters: Optional[Dict[str, Callable[[str], bool]]] = None,
    search_paths: Optional[List[str]] = None,
) -> None:
    """Check the configuration by provided list of validation schemas.

    :param config: Configuration to check
    :param schemas: List of validation schemas
    :param extra_formaters: Additional custom formaters
    :param search_paths: List of paths where to search for the file, defaults to None
    :raises SPSDKError: Invalid validation schema or configuration
    """
    custom_formatters: Dict[str, Callable[[str], bool]] = {
        "dir": lambda x: os.path.isdir(x.replace("\\", "/")),
        "file": lambda x: bool(find_file(x, search_paths=search_paths, raise_exc=False)),
        "file_name": lambda x: os.path.basename(x.replace("\\", "/")) not in ("", None),
        "optional_file": lambda x: not x
        or bool(find_file(x, search_paths=search_paths, raise_exc=False)),
        "number": _is_number,
    }
    if isinstance(config, str):
        config_to_check = load_configuration(config)
        config_dir = os.path.dirname(config)
        if search_paths:
            search_paths.append(config_dir)
        else:
            search_paths = [config_dir]
    else:
        config_to_check = copy.deepcopy(config)

    schema: Dict[str, Any] = {}
    for sch in schemas:
        always_merger.merge(schema, copy.deepcopy(sch))
    validator = None
    formats = always_merger.merge(custom_formatters, extra_formaters or {})
    try:
        if ENABLE_DEBUG:
            validator_code = fastjsonschema.compile_to_code(schema, formats=formats)
            write_file(validator_code, "validator_file.py")
        else:
            validator = fastjsonschema.compile(schema, formats=formats)
    except (TypeError, fastjsonschema.JsonSchemaDefinitionException) as exc:
        raise SPSDKError(f"Invalid validation schema to check config: {str(exc)}") from exc
    try:
        if ENABLE_DEBUG:
            # pylint: disable=import-error,import-outside-toplevel
            import validator_file  # type: ignore

            validator_file.validate(config_to_check, formats)
        else:
            assert validator is not None
            validator(config_to_check)
    except fastjsonschema.JsonSchemaValueException as exc:
        message = _print_validation_fail_reason(exc, formats)
        raise SPSDKError(f"Configuration validation failed: {message}") from exc


class CommentedConfig:
    """Class for generating commented config templates or custom configurations."""

    def __init__(
        self,
        main_title: str,
        schemas: List[Dict[str, Any]],
        values: Optional[Dict[str, Any]] = None,
        note: Optional[str] = None,
        export_template: bool = True,
    ):
        """Constructor for Config templates.

        :param main_title: Main title of final template.
        :param schemas: Main description of final template.
        :param values:
            - for configuration, this is dictionary of values to be saved
            - for schema, this is dictionary of override values (overriding default values)
        :param note: Additional Note after title test.
        :param export_template: True to export schema template; False to export custom configuration
        """
        self.main_title = main_title
        self.schemas = schemas
        self.values = values
        self.indent = 0
        self.note = note
        self.export_template = export_template
        assert export_template or values, "values must be defined for configuration export"

    @staticmethod
    def _get_title_block(title: str, description: Optional[str] = None) -> str:
        """Get unified title blob.

        :param title: Simple title of block
        :param description: Description of block
        :return: ASCII art block
        """
        block_len = 100
        delimiter = "-" * block_len
        title_str = f" == {title} == "
        title_str = title_str.center(block_len)
        descr_list = (
            [text.center(block_len) + "\n" for text in textwrap.wrap(description, block_len)]
            if description
            else [""]
        )

        ret = delimiter + "\n" + title_str + "\n"
        for descr in descr_list:
            ret += descr
        ret += delimiter
        return ret

    @staticmethod
    def get_property_optional_required(key: str, block: Dict[str, Any]) -> str:
        """Function to determine if the config property is required or not.

        :param key: Name of config record
        :param block: Source data block
        :return: Final description.
        """
        schema_kws = ["allOf", "anyOf", "oneOf", "if", "then", "else"]

        def _find_required(d_in: Dict[str, Any]) -> Optional[List[str]]:
            if "required" in d_in:
                return d_in["required"]

            for d_v in d_in.values():
                if isinstance(d_v, dict):
                    ret = _find_required(d_v)
                    if ret:
                        return ret
            return None

        def _find_required_in_schema_kws(schema_node: Union[List, Dict[str, Any]]) -> List[str]:
            """Find all required properties in structure composed of nested properties."""
            all_props: List[str] = []
            if isinstance(schema_node, dict):
                for k, v in schema_node.items():
                    if k == "required":
                        all_props.extend(v)
                    elif k in schema_kws:
                        req_props = _find_required_in_schema_kws(v)
                        all_props.extend(req_props)
            if isinstance(schema_node, list):
                for item in schema_node:
                    req_props = _find_required_in_schema_kws(item)
                    all_props.extend(req_props)
            return list(set(all_props))

        if "required" in block and key in block["required"]:
            return "Required"

        for val in block.values():
            if isinstance(val, dict):
                ret = _find_required(val)
                if ret and key in ret:
                    return "Conditionally required"

        actual_kws = {k: v for k, v in block.items() if k in schema_kws}
        ret = _find_required_in_schema_kws(actual_kws)
        if key in ret:
            return "Conditionally required"

        return "Optional"

    def _create_object_block(
        self,
        block: Dict[str, Dict[str, Any]],
        custom_value: Optional[Union[Dict[str, Any], List[Any]]] = None,
    ) -> Union[CMap, CSeq]:
        """Private function used to create object block with data.

        :param block: Source block with data
        :param custom_value:
            Optional dictionary or List of properties to be exported.
        It is recommended to pass OrderedDict to preserve the key order.
            - key is property ID to be exported
            - value is its value; or None if default value shall be used
        :return: CMap or CSeq base configuration object
        :raises SPSDKError: In case of invalid data pattern.
        """
        assert block.get("type") == "object"
        self.indent += 1

        assert "properties" in block.keys()

        if "oneOf" in block["properties"].keys():
            cfg_s = self._create_one_of_block_array(block["properties"], custom_value)  # type: ignore
            self.indent -= 1
            return cfg_s

        cfg_m = CMap()
        for key in self._get_schema_block_keys(block):
            assert (
                key in block["properties"].keys()
            ), f"Missing key ({key}, in block properties. Block title: {block.get('title', 'Unknown')})"

            # Skip the record in case that custom value key is defined,
            # but it has None value as a mark to not use this record
            value = custom_value.get(key, None) if custom_value else None  # type: ignore
            if not self.export_template and value is None:
                continue

            val_p: Dict = block["properties"][key]
            value_to_add, comment, title = self._get_schema_value(val_p, value)

            cfg_m[key] = value_to_add
            p_required = self.get_property_optional_required(key, block)
            cfg_m.yaml_add_eol_comment(f"[{p_required}]{', ' if comment else ''}{comment}", key=key)
            if title:
                cfg_m.yaml_set_comment_before_after_key(
                    key,
                    self._get_title_block(title),
                    indent=SPSDK_YML_INDENT * (self.indent - 1),
                )
        self.indent -= 1
        return cfg_m

    def _create_array_block(
        self, block: Dict[str, Dict[str, Any]], custom_value: Optional[List[Any]]
    ) -> CSeq:
        """Private function used to create array block with data.

        :param block: Source block with data
        :return: CS base configuration object
        :raises SPSDKError: In case of invalid data pattern.
        """
        assert block.get("type") == "array"
        assert "items" in block.keys()
        self.indent += 1
        val_i: Dict = block["items"]
        if "oneOf" in val_i.keys():
            cfg_s = self._create_one_of_block_array(val_i, custom_value)
        else:
            cfg_s = CSeq()
            if custom_value:
                for cust_val in custom_value:
                    value, _, _ = self._get_schema_value(val_i, cust_val)
                    if isinstance(value, list):
                        cfg_s.extend(value)
                    else:
                        cfg_s.append(value)
            elif self.export_template:
                value, _, _ = self._get_schema_value(val_i, None)
                # the template_value can be the actual list(not only one element)
                if isinstance(value, list):
                    cfg_s.extend(value)
                else:
                    cfg_s.append(value)
        self.indent -= 1
        return cfg_s

    @staticmethod
    def _find_matching_oneof_option(one_of: List[Dict[str, Any]], cust_val: Any) -> Dict[str, Any]:
        """Find matching "oneOf" schema for given custom value.

        :param one_of: list of oneOf schemas
        :param cust_val: custom value; currently must be dictionary
        :raise SPSDKError: if not found
        """
        assert (
            isinstance(cust_val, dict) and cust_val
        ), "currently the implementation supports only dictionary as custom value"
        for option in one_of:
            assert option.get("type") == "object"
            properties = option.get("properties")
            assert properties, "non-empty properties must be defined"
            if all([key in properties for key in cust_val.keys()]):
                return option
        raise SPSDKError(f"for custom value {str(cust_val)}, no corresponding `oneOf` schema found")

    def _create_one_of_block_array(
        self, block: Dict[str, Dict[str, Any]], custom_value: Optional[List[Any]]
    ) -> CSeq:
        """Private function used to create oneOf block with data, and return as an array that contains all values.

        :param block: Source block with data
        :param custom_value: custom value to fill the array
        :return: CS base configuration object
        """
        self.indent += 1
        ret = CSeq()
        one_of = block["oneOf"]
        assert isinstance(one_of, list)
        if custom_value is not None:
            for i, cust_val in enumerate(custom_value):
                option = self._find_matching_oneof_option(one_of, cust_val)
                value, loc_comment, loc_title = self._get_schema_value(option, cust_val)
                ret.append(value)
                if loc_comment:
                    ret.yaml_add_eol_comment(f"{loc_comment}", key=i)
                if loc_title:
                    ret.yaml_set_comment_before_after_key(
                        key=i,
                        before=self._get_title_block(loc_title),
                        indent=SPSDK_YML_INDENT * (self.indent - 1),
                    )
        elif self.export_template:
            option_types = ",".join([str(x.get("type")) for x in one_of])
            title = f"List of possible {len(one_of)} options. Option types[{option_types}]"
            for i, option in enumerate(one_of):
                value, loc_comment, loc_title = self._get_schema_value(option, None)
                ret.append(value)
                ret.yaml_add_eol_comment(
                    f"[Example of possible configuration #{i}] {loc_comment}", key=i
                )
                if loc_title:
                    ret.yaml_set_comment_before_after_key(
                        key=i,
                        before=self._get_title_block(loc_title),
                        indent=SPSDK_YML_INDENT * (self.indent - 1),
                    )
            ret.yaml_set_comment_before_after_key(
                key=0,
                before=self._get_title_block(title),
                indent=SPSDK_YML_INDENT * (self.indent - 1),
            )
        self.indent -= 1
        return ret

    def _get_schema_value(
        self, block: Dict[str, Any], custom_value: Any
    ) -> Tuple[Optional[Union[CMap, CSeq, str, int, float, List]], Optional[str], Optional[str]]:
        """Private function used to fill up configuration block with data.

        :param block: Source block with data
        :param custom_value: value to be saved instead of default value
        :return: CM/CS base configuration object with comment
        :raises SPSDKError: In case of invalid data pattern.
        """
        title = None
        if "oneOf" in block.keys():
            ret = self._create_one_of_block_array(block, custom_value)
        else:
            schema_type = block.get("type")
            if not schema_type:
                raise SPSDKError(f"Type not available in block: {block}")
            assert schema_type, f"Type not available in block: {block}"

            if schema_type == "object":
                assert (custom_value is None) or isinstance(custom_value, dict)
                ret = self._create_object_block(block, custom_value)  # type: ignore
            elif schema_type == "array":
                assert (custom_value is None) or isinstance(custom_value, list)
                ret = self._create_array_block(block, custom_value)  # type: ignore
            else:
                assert "template_value" in block.keys()
                ret = custom_value if (custom_value is not None) else block.get("template_value")

        assert isinstance(ret, (CMap, CSeq, str, int, float, list)) or (ret is None)

        p_title = block.get("title", "")
        p_descr = block.get("description", "")
        p_enum = block.get("enum", None)
        if "template_title" in block.keys():
            title = block["template_title"]
        if p_enum:
            if p_descr.endswith("."):
                p_descr = p_descr[:-1]
            p_descr = f"{p_descr}{'; ' if p_descr else ''}Possible options:{p_enum}"
        if p_title.endswith("."):
            p_title = p_title[:-1]
        if p_descr.lower().startswith(p_title.lower()):
            comment = p_descr
        else:
            comment = f"{p_title}{'; ' if p_descr else ''}{p_descr}"
        return ret, comment, title

    @staticmethod
    def _get_schema_block_keys(schema: Dict[str, Dict[str, Any]]) -> List[str]:
        """Creates list of property keys in given schema.

        :param schema: Input schema piece.
        :return: List of all property keys.
        """
        if "properties" not in schema:
            return []
        return [
            key
            for key in schema["properties"]
            if "skip_in_template" not in schema["properties"][key]
            or schema["properties"][key]["skip_in_template"] is False
        ]

    def export(self) -> CMap:
        """Export configuration template into CommentedMap.

        :raises SPSDKError: Error
        :return: Configuration template in CM.
        """
        loc_schemas = copy.deepcopy(self.schemas)
        # 1. Get blocks with their titles and lists of their keys
        block_list: Dict[str, Any] = {}
        for schema in loc_schemas:
            title = schema.get("title", "General Options")
            if title in block_list:
                property_list = block_list[title]["properties"]
                assert isinstance(property_list, list)
                property_list.extend(
                    [
                        x
                        for x in self._get_schema_block_keys(schema)
                        if x not in block_list[title]["properties"]
                    ]
                )
            else:
                block_list[title] = {}
                block_list[title]["properties"] = self._get_schema_block_keys(schema)
                block_list[title]["description"] = schema.get("description", "")

        # 2. Merge all schemas together to get whole single schema
        schemas_merger = SPSDKMerger(
            [(list, ["set"]), (dict, ["merge"]), (set, ["union"])],
            ["override"],
            ["override"],
        )

        merged: Dict[str, Any] = {}
        for schema in loc_schemas:
            schemas_merger.merge(merged, copy.deepcopy(schema))

        # 3. Create order of individual settings

        order_dict: Dict[str, Any] = OrderedDict()
        properties_for_template = self._get_schema_block_keys(merged)
        for block in block_list.values():
            block_properties: list = block["properties"]
            # block_properties.sort()
            for block_property in block_properties:
                if block_property in properties_for_template:
                    order_dict[block_property] = merged["properties"][block_property]
        merged["properties"] = order_dict

        try:
            self.indent = 0
            # 4. Go through all individual logic blocks
            cfg = self._create_object_block(merged, self.values)
            assert isinstance(cfg, CMap)
            # 5. Add main title of configuration
            title = f"===========  {self.main_title}  ===========\n"
            if self.note:
                title += f"\n {'-'*50} Note {'-'*50}\n{self.note}\n\n"
            cfg.yaml_set_start_comment(title)
            for title, info in block_list.items():
                description = info["description"]
                assert isinstance(description, str) or description is None
                cfg.yaml_set_comment_before_after_key(
                    info["properties"][0], self._get_title_block(title, description)
                )
            return cfg

        except Exception as exc:
            raise SPSDKError(f"Template generation failed: {str(exc)}") from exc

    def export_to_yaml(self) -> str:
        """Export Configuration template directly into YAML string format.

        :return: YAML string.
        """
        return self.convert_cm_to_yaml(self.export())

    @staticmethod
    def convert_cm_to_yaml(config: CMap) -> str:
        """Convert Commented Map for into final YAML string.

        :param config: Configuration in CM format.
        :return: YAML string with configuration to use to store in file.
        """
        yaml = YAML(pure=True)
        yaml.indent(sequence=SPSDK_YML_INDENT * 2, offset=SPSDK_YML_INDENT)
        stream = io.StringIO()
        yaml.dump(config, stream)
        yaml_data = stream.getvalue()

        return yaml_data


class ValidationSchemas:
    """Manager for validation schemas."""

    _instancies: Dict[str, Dict[str, Any]] = {}

    @staticmethod
    def get_schema_file(sch_file: str) -> Dict[str, Any]:
        """Return load schema file. Use SingleTon behavior.

        :param sch_file: Path to schema config file.
        :raises SPSDKError: Invalid schema config file.
        :return: Loaded schema file.
        """
        abs_path = os.path.abspath(sch_file)
        if abs_path not in ValidationSchemas._instancies:
            try:
                schema_cfg = load_configuration(abs_path)
            except (FileNotFoundError, YAMLError, UnicodeDecodeError) as exc:
                raise SPSDKError("Invalid validation scheme configuration file.") from exc
            ValidationSchemas._instancies[abs_path] = schema_cfg

        return ValidationSchemas._instancies[abs_path]


class ConfigTemplate(CommentedConfig):
    """Deprecated, kept for backward compatibility only."""



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Boot Selection for SB file."""

from typing import List, Sequence

from spsdk.utils.crypto.abstract import BaseClass
from spsdk.utils.misc import DebugInfo

from ..misc import SecBootBlckSize
from ..sb2.commands import (
    CmdBaseClass,
    CmdCall,
    CmdErase,
    CmdFill,
    CmdJump,
    CmdLoad,
    CmdMemEnable,
    CmdNop,
    CmdProg,
    CmdReset,
)
from .commands import parse_v1_command
from .headers import BootSectionHeaderV1, SecureBootFlagsV1


class BootSectionV1(BaseClass):
    """Boot Section for SB file 1.x."""

    def __init__(self, section_id: int, flags: SecureBootFlagsV1 = SecureBootFlagsV1.NONE):
        """Initialize BootSectionV1.

        :param section_id: unique section ID, 32-bit int
        :param flags: see SecureBootFlagsV1
        """
        self._header = BootSectionHeaderV1(section_id, flags)
        self._commands: List[CmdBaseClass] = []

    @property
    def section_id(self) -> int:
        """Return unique ID of the section, 32 number."""
        return self._header.section_id

    @property
    def flags(self) -> SecureBootFlagsV1:
        """Return section flags."""
        return self._header.flags

    @property
    def bootable(self) -> bool:
        """Return whether section is bootable."""
        return self._header.bootable

    @property
    def rom_last_tag(self) -> bool:
        """ReturnROM_LAST_TAG flag.

        The last section header in an image always has its ROM_LAST_TAG flag set to help the ROM know at what point
        to stop searching.
        """
        return self._header.rom_last_tag

    @rom_last_tag.setter
    def rom_last_tag(self, value: bool) -> None:
        """Setter.

        :param value: ROM_LAST_TAG flag
        """
        self._header.rom_last_tag = value

    @property
    def cmd_size(self) -> int:
        """Return size of the binary representation of the commands."""
        return sum(cmd.raw_size for cmd in self._commands)

    @property
    def size(self) -> int:
        """Return size of the binary representation of the section in bytes."""
        result = self._header.raw_size + self.cmd_size
        return result

    def info(self) -> str:
        """Return string representation."""
        result = "[BootSection-V1]\n"
        result += f"ID: {self._header.section_id}\n"
        result += f"NumBlocks: {self._header.num_blocks}\n"
        result += self._header.info() + "\n"
        result += "[BootSection-commands]\n"
        for cmd in self._commands:
            result += cmd.info()
        return result

    @property
    def commands(self) -> Sequence[CmdBaseClass]:
        """Return sequence of all commands in the section."""
        return self._commands

    def append(self, cmd: CmdBaseClass) -> None:
        """Append command.

        :param cmd: to be added
        """
        assert isinstance(
            cmd,
            (
                CmdNop,
                CmdErase,
                CmdLoad,
                CmdFill,
                CmdJump,
                CmdCall,
                CmdReset,
                CmdMemEnable,
                CmdProg,
            ),
        )
        self._commands.append(cmd)

    def update(self) -> None:
        """Update settings."""
        self._header.num_blocks = SecBootBlckSize.to_num_blocks(self.cmd_size)

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Return binary representation of the class (serialization)."""
        self.update()
        dbg_info.append_section("Section")
        data = self._header.export()
        dbg_info.append_binary_data("Section-header", data)
        dbg_info.append_section("Commands")
        for cmd in self._commands:
            cmd_data = cmd.export(dbg_info)
            data += cmd_data
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "BootSectionV1":
        """Deserialization from binary format.

        :param data: to be parsed
        :param offset: to start parsing
        :return: the parsed instance
        """
        header = BootSectionHeaderV1.parse(data, offset)
        result = BootSectionV1(0)
        result._header = header
        # commands
        cmd_base = offset + header.raw_size
        cmd_ofs = 0
        end_ofs = result._header.num_blocks * SecBootBlckSize.BLOCK_SIZE
        while cmd_ofs < end_ofs:
            cmd = parse_v1_command(data, cmd_base + cmd_ofs)
            result.append(cmd)
            cmd_ofs += cmd.raw_size
        return result



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Trust provisioning."""
import os

from spsdk import SPSDK_DATA_FOLDER

TP_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "tp")
TP_SCH_FILE: str = os.path.join(TP_DATA_FOLDER, "sch_tp.yaml")

from .exceptions import SPSDKTpConfigError, SPSDKTpError, SPSDKTpTargetError, SPSDKTpTimeoutError
from .tp_intf import TpDevInterface, TpIntfDescription, TpTargetInterface
from .tpconfig import TrustProvisioningConfig
from .tphost import TrustProvisioningHost
from .utils import (
    get_supported_devices,
    get_tp_device_class,
    get_tp_device_types,
    get_tp_target_class,
    get_tp_target_types,
    scan_tp_devices,
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing SBFile version 1."""

from spsdk.mboot import ExtMemId

from ..misc import BcdVersion3
from ..sb2 import commands
from ..sb2.commands import (
    CmdCall,
    CmdErase,
    CmdFill,
    CmdJump,
    CmdLoad,
    CmdMemEnable,
    CmdNop,
    CmdProg,
    CmdReset,
    CmdTag,
)
from .headers import SecureBootFlagsV1
from .images import SecureBootV1
from .sections import BootSectionV1

__all__ = [
    # main classes
    "SecureBootV1",
    "BootSectionV1",
    # commands
    "CmdFill",
    "CmdNop",
    "CmdTag",
    "CmdMemEnable",
    "CmdErase",
    "CmdReset",
    "CmdCall",
    "CmdLoad",
    "CmdJump",
    "CmdProg",
    # helper classes and enums
    "SecureBootFlagsV1",
    "BcdVersion3",
    "ExtMemId",
]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""SerialProxy serves as patch replacement for serial.Serial class."""

import logging

# pylint: disable=unused-import  # Type is necessary for Mypy
from typing import Dict, Optional, Type

logger = logging.getLogger(__name__)


class SerialProxy:
    """SerialProxy is used to simulate communication with serial device.

    It can be used as mock.patch for serial.Serial class.
    @patch(<your.package>.Serial, SerialProxy.init_proxy(pre_recorded_responses))
    """

    responses: Dict[bytes, bytes] = {}
    ignore_ack: bool = False

    @classmethod
    def init_proxy(cls, data: Dict[bytes, bytes], ignore_ack: bool = False) -> "Type[SerialProxy]":
        """Initialized response dictionary of write and read bytes.

        :param data: Dictionary of write and read bytes
        :param ignore_ack: Don't modify internal buffer upon receiving a ACK packet
        :return: SerialProxy class with configured data
        """
        cls.responses = data
        cls.ignore_ack = ignore_ack
        return cls

    def __init__(self, port: str, timeout: int, baudrate: int, write_timeout: Optional[int] = None):
        """Basic initialization for serial.Serial class.

        __init__ signature must accommodate instantiation of serial.Serial

        :param port: Serial port name
        :param timeout: timeout (does nothing)
        :param write_timeout: does nothing
        :param baudrate: Serial port speed (does nothing)
        """
        self.port = port
        self.timeout = timeout
        self.write_timeout = write_timeout
        self.baudrate = baudrate
        self.is_open = False
        self.buffer = bytes()

    def open(self) -> None:
        """Simulates opening a serial port."""
        self.is_open = True

    def close(self) -> None:
        """Simulates closing a serial port."""
        self.is_open = False

    def write(self, data: bytes) -> None:
        """Simulates a write, currently just pick up response from responses.

        :param data: Bytes to write, key in responses
        """
        logger.debug(f"I got: {data!r}")
        if self.ignore_ack and data == b"\x5a\xa1":
            logger.debug("ACK received and ignored")
            return
        self.buffer = self.responses[data]
        logger.debug(f"setting buffer to: '{self.buffer!r}'")

    def read(self, length: int) -> bytes:
        """Read portion of pre-configured data.

        :param length: Amount of data to read from buffer
        :return: Data read
        """
        segment = self.buffer[:length]
        self.buffer = self.buffer[length:]
        logger.debug(f"I responded with: '{segment!r}'")
        return segment

    def info(self) -> str:
        """Text information about the interface."""
        return self.__class__.__name__

    def reset_input_buffer(self) -> None:
        """Simulates resetting input buffer."""

    def reset_output_buffer(self) -> None:
        """Simulates resetting output buffer."""

    def flush(self) -> None:
        """Simulates flushing input buffer."""


class SimpleReadSerialProxy(SerialProxy):
    """SimpleReadSerialProxy is used to simulate communication with serial device.

    It simplifies reading method.
    @patch(<your.package>.Serial, SerialProxy.init_proxy(pre_recorded_responses))
    """

    FULL_BUFFER = bytes()

    @classmethod
    def init_data_proxy(cls, data: bytes) -> "Type[SimpleReadSerialProxy]":
        """Initialized response dictionary of write and read bytes.

        :param data: Dictionary of write and read bytes
        :return: SerialProxy class with configured data
        """
        cls.FULL_BUFFER = data
        return cls

    def __init__(self, port: str, timeout: int, baudrate: int, write_timeout: Optional[int] = None):
        """Basic initialization for serial.Serial class.

        __init__ signature must accommodate instantiation of serial.Serial

        :param port: Serial port name
        :param timeout: timeout (does nothing)
        :param write_timeout: does nothing
        :param baudrate: Serial port speed (does nothing)
        """
        super().__init__(port=port, timeout=timeout, baudrate=baudrate, write_timeout=write_timeout)
        self.buffer = self.FULL_BUFFER

    def write(self, data: bytes) -> None:
        """Simulates a write method, but it does nothing.

        :param data: Bytes to write, key in responses
        """



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Trust provisioning adapters."""

from typing import Dict, Type

from .. import TpDevInterface, TpTargetInterface
from .tpdev_scard import TpDevSmartCard
from .tptarget_blhost import TpTargetBlHost

# Dict mapping TP device name to its adapter
TP_DEVICES: Dict[str, Type[TpDevInterface]] = {
    TpDevSmartCard.NAME: TpDevSmartCard,
}

# Dict mapping TP target name to its adapter
TP_TARGETS: Dict[str, Type[TpTargetInterface]] = {
    TpTargetBlHost.NAME: TpTargetBlHost,
}

try:
    # Import TP Device model if present in this build
    from .tpdev_model import TpDevSwModel

    TP_DEVICES.update({TpDevSwModel.NAME: TpDevSwModel})
except ImportError:
    pass


try:
    # Import TP Target model if present in this build
    from .tptarget_model import TpTargetSwModel

    TP_TARGETS.update({TpTargetSwModel.NAME: TpTargetSwModel})
except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains Bootable image related code."""

import abc
import logging
import os
import re
import sys
from copy import deepcopy
from enum import Enum
from typing import Any, Dict, List, Optional, Type

from spsdk.exceptions import SPSDKKeyError, SPSDKParsingError, SPSDKTypeError, SPSDKValueError
from spsdk.image.bootable_image import BIMG_DATABASE_FILE, BIMG_SCH_FILE
from spsdk.image.fcb.fcb import FCB
from spsdk.image.segments import FlexSPIConfBlockFCB, XMCDHeader
from spsdk.image.xmcd.xmcd import XMCD, ConfigurationBlockType, MemoryType
from spsdk.utils.database import Database
from spsdk.utils.images import BinaryImage, BinaryPattern
from spsdk.utils.misc import load_binary, write_file
from spsdk.utils.schema_validator import ConfigTemplate, ValidationSchemas, check_config

logger = logging.getLogger(__name__)

BIMG_CLASSES = [
    "BootableImageRtxxx",
    "BootableImageLpc55s3xRw61x",
    "BootableImageRt101x",
    "BootableImageRt10xx",
    "BootableImageRt11xx",
    "BootableImageRt118x",
]


class EnumBimgSegments(str, Enum):
    """Bootable image segments."""

    KEYBLOB = "keyblob"
    FCB = "fcb"
    IMAGE_VERSION = "image_version"
    KEYSTORE = "keystore"
    APPLICATION = "application"
    BEE_HEADER_0 = "bee_header_0"
    BEE_HEADER_1 = "bee_header_1"
    XMCD = "xmcd"
    HAB_CONTAINER = "hab_container"
    AHAB_CONTAINER = "ahab_container"

    @staticmethod
    def get_segment_type(segment: "EnumBimgSegments") -> Type:
        """Get the output value type.

        :param segment: Bootable image segment
        :return: True is the segment output type is binary. False otherwise.
        """
        if segment == EnumBimgSegments.IMAGE_VERSION:
            return int
        return bytes


def convert_segment_value(value: bytes, output_type: Type) -> Any:
    """Convert value of segment into desired type.

    :param value: Segment value as bytes.
    :param output_type: Output type.
    :raises SPSDKTypeError: If given type is not supported
    :return: Converted value.
    """
    if output_type == int:
        return int.from_bytes(value, "little")
    if output_type == bytes:
        return value
    raise SPSDKTypeError(f"Unsupported output type {output_type}")


def get_bimg_class(family: str) -> Type["BootableImage"]:
    """Get the class that supports the family.

    :param family: Chip family
    :return: Bootable Image class.
    :raises SPSDKValueError: Invalid family.
    """
    for cls_name in BIMG_CLASSES:
        cls: Type["BootableImage"] = getattr(sys.modules[__name__], cls_name)
        if family in cls.get_supported_families():
            return cls
    raise SPSDKValueError(f"Unsupported family({family}) by Bootable Image.")


class BootableImage:
    """Bootable Image class."""

    IMAGE_PATTERN = BinaryPattern("zeros")

    def __init__(self, family: str, mem_type: str, revision: str = "latest") -> None:
        """Bootable Image constructor.

        :param family: Chip family.
        :param mem_type: Used memory type.
        :param revision: Chip silicon revision.
        :raises SPSDKValueError: Invalid family.
        """
        if family not in self.get_supported_families():
            raise SPSDKValueError(f"Unsupported family: {family}")
        self.family = family
        self.revision = revision
        self.mem_type = mem_type
        self.database = Database(BIMG_DATABASE_FILE)
        self.mem_types: Dict = self.database.get_device_value("mem_types", family, revision)
        if mem_type not in self.mem_types.keys():
            raise SPSDKValueError(f"Unsupported memory type: {mem_type}")
        self.bimg_descr: Dict = self.mem_types[self.mem_type]

    @staticmethod
    @abc.abstractmethod
    def get_segments(mem_type: str) -> List[EnumBimgSegments]:
        """Get list of image segments.

        :param mem_type: Used memory type.
        """

    def get_segment_size(self, segment: EnumBimgSegments) -> int:
        """Get size of segment. Return -1 if it is the last segment."""
        segments = {
            EnumBimgSegments.FCB: 512,
            EnumBimgSegments.BEE_HEADER_0: 512,
            EnumBimgSegments.BEE_HEADER_1: 512,
            EnumBimgSegments.KEYBLOB: 256,
            EnumBimgSegments.KEYSTORE: 2048,
            EnumBimgSegments.IMAGE_VERSION: 4,
            EnumBimgSegments.APPLICATION: -1,
            EnumBimgSegments.HAB_CONTAINER: -1,
            EnumBimgSegments.AHAB_CONTAINER: -1,
        }
        return segments[segment]

    def parse(self, binary: bytes) -> None:
        """Parse binary into bootable image object.

        :param binary: Full binary of bootable image.
        """
        for segment in self.get_segments(self.mem_type):
            offset_from = self.bimg_descr[f"{segment.value}_offset"]
            offset_to = (
                offset_from + self.get_segment_size(segment)
                if self.get_segment_size(segment) != -1
                else None
            )
            value = binary[offset_from:offset_to]
            setattr(self, segment.value, value)

    @staticmethod
    def get_validation_schemas(
        family: str, mem_type: str, revision: str = "latest"
    ) -> List[Dict[str, Any]]:
        """Get validation schema for the family.

        :param family: Chip family
        :param mem_type: Used memory type.
        :param revision: Chip revision specification, as default, latest is used.
        :return: List of validation schema dictionaries.
        """
        bimg_cls = get_bimg_class(family)
        sch_cfg = deepcopy(ValidationSchemas.get_schema_file(BIMG_SCH_FILE))
        sch_cfg["family_rev"]["properties"]["family"]["enum"] = bimg_cls.get_supported_families()
        sch_cfg["family_rev"]["properties"]["revision"]["enum"] = bimg_cls.get_supported_revisions(
            family
        )

        sch_cfg["family_rev"]["properties"]["memory_type"][
            "enum"
        ] = bimg_cls.get_supported_memory_types(family, revision)
        schemas = [sch_cfg["family_rev"]]
        for segment in bimg_cls.get_segments(mem_type):
            schemas.append(sch_cfg[segment.value])
        return schemas

    def store_config(self, output: str) -> None:
        """Store bootable image into configuration and binary blocks.

        :param output: Path to output folder to store bootable image configuration.
        """
        schemas = self._get_validation_schemas()
        override: Dict[str, str] = {}
        override["family"] = self.family
        override["revision"] = self.revision
        override["memory_type"] = self.mem_type
        for segment in self.get_segments(self.mem_type):
            segment_value = getattr(self, segment.value)
            if segment_value is not None:
                segment_value = convert_segment_value(
                    segment_value, EnumBimgSegments.get_segment_type(segment)
                )
                if isinstance(segment_value, bytes):
                    override[segment.value] = f"{segment.value}.bin"
                else:
                    override[segment.value] = segment_value
            else:
                override[segment.value] = ""

        config = ConfigTemplate(
            f"Bootable Image Configuration for {self.family}.",
            schemas,
            override,
        ).export_to_yaml()
        write_file(
            config,
            os.path.join(output, f"bootable_image_{self.family}_{self.mem_type}.yaml"),
        )
        for segment in self.get_segments(self.mem_type):
            segment_value = getattr(self, segment.value)
            if segment_value is not None:
                if EnumBimgSegments.get_segment_type(segment) == bytes:
                    write_file(
                        segment_value, os.path.join(output, f"{segment.value}.bin"), mode="wb"
                    )

    @classmethod
    def load_from_config(
        cls, config: Dict, search_paths: Optional[List[str]] = None
    ) -> "BootableImage":
        """Load bootable image from configuration.

        :param config: Configuration of Bootable image.
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        bimg_cls = get_bimg_class(config["family"])
        check_config(config, cls.get_validation_schemas_family())
        chip_family = config["family"]
        mem_type = config["memory_type"]
        revision = config.get("revision", "latest")
        schemas = cls.get_validation_schemas(chip_family, mem_type, revision)
        check_config(config, schemas, search_paths=search_paths)
        params = {"family": chip_family, "mem_type": mem_type, "revision": revision}
        for segment in bimg_cls.get_segments(mem_type):
            segment_value = config.get(segment.value)
            if EnumBimgSegments.get_segment_type(segment) == bytes:
                value = (
                    load_binary(segment_value, search_paths=search_paths) if segment_value else None
                )
            else:
                value = segment_value
            params[segment.value] = value
        return bimg_cls(**params)

    def image_info(self) -> BinaryImage:
        """Create Binary image of bootable image.

        :return: BinaryImage object of bootable image.
        """
        bin_image = BinaryImage(
            name=f"Bootable Image for {self.family}", size=0, pattern=self.IMAGE_PATTERN
        )
        for segment in self.get_segments(self.mem_type):
            segment_value = getattr(self, segment.value)
            if segment_value:
                bin_image.add_image(
                    BinaryImage(
                        name=segment.value,
                        size=len(segment_value),
                        offset=self.bimg_descr[f"{segment.value}_offset"],
                        binary=segment_value,
                        parent=bin_image,
                    )
                )
        return bin_image

    def export(self) -> bytes:
        """Export bootable image.

        :return: Complete binary of bootable image.
        """
        return self.image_info().export()

    @staticmethod
    def get_validation_schemas_family() -> List[Dict[str, Any]]:
        """Create the validation schema just for supported families.

        :return: List of validation schemas for Bootable Image supported families.
        """
        sch_cfg = ValidationSchemas.get_schema_file(BIMG_SCH_FILE)
        return [sch_cfg["family_rev"]]

    def _get_validation_schemas(self) -> List[Dict[str, Any]]:
        """Get validation schema.

        :return: List of validation schema dictionaries.
        """
        return self.get_validation_schemas(self.family, self.mem_type, self.revision)

    @staticmethod
    def generate_config_template(family: str, mem_type: str, revision: str = "latest") -> str:
        """Get validation schema for the family.

        :param family: Chip family
        :param mem_type: Used memory type.
        :param revision: Chip revision specification, as default, latest is used.
        :return: Validation schema.
        """
        schemas = BootableImage.get_validation_schemas(family, mem_type, revision)
        override = {}
        override["family"] = family
        override["revision"] = revision
        override["memory_type"] = mem_type

        return ConfigTemplate(
            f"Bootable Image Configuration template for {family}.",
            schemas,
            override,
        ).export_to_yaml()

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get list of all supported families by bootable image.

        :return: List of families.
        """
        return Database(BIMG_DATABASE_FILE).devices.device_names

    @staticmethod
    def get_supported_memory_types(family: str, revision: str = "latest") -> List[str]:
        """Return list of supported memory types.

        :return: List of supported families.
        """
        database = Database(BIMG_DATABASE_FILE)
        return list(database.get_device_value("mem_types", family, revision).keys())

    @staticmethod
    def get_memory_type_config(
        family: str, mem_type: str, revision: str = "latest"
    ) -> Dict[str, Any]:
        """Return dictionary with configuration for specific memory type.

        :raises SPSDKKeyError: If memory type does not exist in database
        :return: Dictionary with configuration.
        """
        if mem_type not in BootableImage.get_supported_memory_types(family):
            raise SPSDKKeyError(f"Memory type not supported: {mem_type}")
        database = Database(BIMG_DATABASE_FILE)
        mem_types: Dict = database.get_device_value("mem_types", family, revision)
        return mem_types[mem_type]

    @staticmethod
    def get_supported_revisions(family: str) -> List[str]:
        """Return list of supported revisions.

        :return: List of supported revisions.
        """
        database = Database(BIMG_DATABASE_FILE)
        revisions = ["latest"]
        revisions.extend(database.devices.get_by_name(family).revisions.revision_names)
        return revisions

    @classmethod
    def _load_bin_from_config(
        cls, config: Dict, config_key: str, search_paths: Optional[List[str]] = None
    ) -> Optional[bytes]:
        """Load the binary defined in condig file."""
        bin_path = config.get(config_key)
        if not bin_path or bin_path == "":
            return None
        return load_binary(bin_path, search_paths=search_paths)


class BootableImageRtxxx(BootableImage):
    """Bootable Image class for RTxxx devices."""

    def __init__(
        self,
        family: str,
        mem_type: str,
        revision: str = "latest",
        keyblob: Optional[bytes] = None,
        fcb: Optional[bytes] = None,
        image_version: Optional[int] = None,
        keystore: Optional[bytes] = None,
        application: Optional[bytes] = None,
    ) -> None:
        """Bootable Image constructor for RTxxx devices.

        :param keyblob: Key Blob block, defaults to None
        :param fcb: FCB block, defaults to None
        :param image_version: Image version number, defaults to 0
        :param keystore: Key store block, defaults to None
        :param app: Application block, defaults to None
        """
        super().__init__(family, mem_type, revision)
        self.keyblob = keyblob
        self.fcb = fcb
        if self.fcb:
            self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
            self.fcb_obj.parse(self.fcb)

        self.image_version = (image_version or 0).to_bytes(4, "little")
        self.keystore = keystore
        self.application = application

    @staticmethod
    def get_segments(mem_type: str) -> List[EnumBimgSegments]:
        """Get list of image segments.

        :param mem_type: Used memory type.
        """
        return [
            EnumBimgSegments.KEYBLOB,
            EnumBimgSegments.FCB,
            EnumBimgSegments.IMAGE_VERSION,
            EnumBimgSegments.KEYSTORE,
            EnumBimgSegments.APPLICATION,
        ]

    def parse(self, binary: bytes) -> None:
        """Parse binary into bootable image object.

        :param binary: Complete binary of bootable image.
        """
        # first of all we need to identify where the image starts.
        # That could be determined by FCB block that start at zero offset
        # as some compilers do that, otherwise we assume standard start at zero offset
        start_block_offset = 0
        if binary[:4] == FlexSPIConfBlockFCB.TAG:
            start_block_offset = self.bimg_descr["fcb_offset"]

        # KeyBlob
        if start_block_offset == 0:
            offset = self.bimg_descr["keyblob_offset"]
            self.keyblob = binary[offset : offset + self.get_segment_size(EnumBimgSegments.KEYBLOB)]
        else:
            self.keyblob = None
        # FCB
        offset = self.bimg_descr["fcb_offset"] - start_block_offset
        self.fcb = binary[offset : offset + self.get_segment_size(EnumBimgSegments.FCB)]
        self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
        self.fcb_obj.parse(self.fcb)
        # Image version
        offset = self.bimg_descr["image_version_offset"] - start_block_offset
        self.image_version = binary[
            offset : offset + self.get_segment_size(EnumBimgSegments.IMAGE_VERSION)
        ]
        # KeyStore
        offset = self.bimg_descr["keystore_offset"] - start_block_offset
        self.keystore = binary[offset : offset + self.get_segment_size(EnumBimgSegments.KEYSTORE)]
        # application
        offset = self.bimg_descr["application_offset"] - start_block_offset
        self.application = binary[offset:]

    def store_config(self, output: str) -> None:
        """Store bootable image into configuration and binary blocks.

        :param output: Path to output folder to store bootable image configuration.
        """
        super().store_config(output)
        if self.fcb_obj:
            write_file(self.fcb_obj.create_config(), os.path.join(output, "fcb.yaml"))

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get list of all supported families by bootable image.

        :return: List of families.
        """
        full_list = BootableImage.get_supported_families()
        # filter out just RTxxx
        return [x for x in full_list if re.match(r"[rR][tT][\dxX]{3}$", x)]


class BootableImageLpc55s3xRw61x(BootableImage):
    """Bootable Image class for LPC55S3x and Rw61x devices."""

    IMAGE_PATTERN = BinaryPattern("ones")

    def __init__(
        self,
        family: str,
        mem_type: str = "flexspi_nor",
        revision: str = "latest",
        fcb: Optional[bytes] = None,
        image_version: Optional[int] = None,
        application: Optional[bytes] = None,
    ) -> None:
        """Bootable Image constructor for Lpc55s3x and Rw61x devices.

        :param mem_type: Used memory type.
        :param fcb: FCB block, defaults to None
        :param image_version: Image version number, defaults to 0
        :param app: Application block, defaults to None
        """
        assert mem_type == "flexspi_nor"
        super().__init__(family, mem_type, revision)
        self.fcb = fcb
        if self.fcb:
            self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
            self.fcb_obj.parse(self.fcb)

        image_version = (image_version or 0) & 0xFFFF
        image_version |= (image_version ^ 0xFFFF) << 16
        self.image_version = image_version.to_bytes(4, "little")
        self.application = application

    @staticmethod
    def get_segments(mem_type: str) -> List[EnumBimgSegments]:
        """Get list of image segments.

        :param mem_type: Used memory type.
        """
        return [
            EnumBimgSegments.FCB,
            EnumBimgSegments.IMAGE_VERSION,
            EnumBimgSegments.APPLICATION,
        ]

    def parse(self, binary: bytes) -> None:
        """Parse binary into bootable image object.

        :param binary: Complete binary of bootable image.
        :raises SPSDKValueError: In case of invalid SW image version.
        """
        # first of all we need to identify where the image starts.
        # That could be determined by FCB block that start at zero offset
        # as some compilers do that, otherwise we assume standard start at zero offset
        start_block_offset = 0
        if binary[:4] == FlexSPIConfBlockFCB.TAG:
            start_block_offset = self.bimg_descr["fcb_offset"]

        # FCB
        offset = self.bimg_descr["fcb_offset"] - start_block_offset
        self.fcb = binary[offset : offset + self.get_segment_size(EnumBimgSegments.FCB)]
        self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
        self.fcb_obj.parse(self.fcb)
        # Image version
        offset = self.bimg_descr["image_version_offset"] - start_block_offset
        image_version = int.from_bytes(
            binary[offset : offset + self.get_segment_size(EnumBimgSegments.IMAGE_VERSION)],
            "little",
        )
        if image_version != 0xFFFFFFFF and (
            image_version & 0xFFFF != ((image_version >> 16) ^ 0xFFFF) & 0xFFFF
        ):
            raise SPSDKValueError("Invalid Image version loaded during parse of bootable image.")
        self.image_version = (image_version & 0xFFFF).to_bytes(4, "little")
        # application
        offset = self.bimg_descr["application_offset"] - start_block_offset
        self.application = binary[offset:]

    def store_config(self, output: str) -> None:
        """Store bootable image into configuration and binary blocks.

        :param output: Path to output folder to store bootable image configuration.
        """
        super().store_config(output)
        if self.fcb_obj:
            write_file(self.fcb_obj.create_config(), os.path.join(output, "fcb.yaml"))

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get list of all supported families by bootable image.

        :return: List of families.
        """
        full_list = BootableImage.get_supported_families()
        # filter out just LPC55S3x and Rw61x
        return [
            x
            for x in full_list
            if re.match(r"[lL][pP][cC]55[sS]3[\dxX]$", x) or re.match(r"[rR][wW]61[\dxX]$", x)
        ]


class BootableImageRt101x(BootableImage):
    """Bootable Image class for RT11x devices."""

    def __init__(
        self,
        family: str,
        mem_type: str,
        revision: str = "latest",
        keyblob: Optional[bytes] = None,
        fcb: Optional[bytes] = None,
        hab_container: Optional[bytes] = None,
    ) -> None:
        """Bootable Image constructor for RT1xxx devices.

        :param family: Chip family.
        :param mem_type: Used memory type.
        :param revision: Chip silicon revision.
        :param keyblob: BEE encryption header 0
        :param fcb: FCB block, defaults to None
        :param hab_container: Boot image container
        """
        super().__init__(family, mem_type, revision)
        self.keyblob = keyblob
        self.fcb = fcb
        if self.fcb:
            self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
            self.fcb_obj.parse(self.fcb)
        self.hab_container = hab_container

    @staticmethod
    def get_segments(mem_type: str) -> List[EnumBimgSegments]:
        """Get list of image segments.

        :param mem_type: Used memory type.
        """
        return [
            EnumBimgSegments.KEYBLOB,
            EnumBimgSegments.FCB,
            EnumBimgSegments.HAB_CONTAINER,
        ]

    def parse(self, binary: bytes) -> None:
        """Parse binary into bootable image object.

        :param binary: Full binary of bootable image.
        """
        super().parse(binary)
        if self.fcb:
            self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
            self.fcb_obj.parse(self.fcb)

    def store_config(self, output: str) -> None:
        """Store bootable image into configuration and binary blocks.

        :param output: Path to output folder to store bootable image configuration.
        """
        super().store_config(output)
        if self.fcb_obj:
            write_file(self.fcb_obj.create_config(), os.path.join(output, "fcb.yaml"))

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get list of all supported families by bootable image.

        :return: List of families.
        """
        return ["rt101x"]


class BootableImageRt10xx(BootableImage):
    """Bootable Image class for RT1xxx devices."""

    def __init__(
        self,
        family: str,
        mem_type: str,
        revision: str = "latest",
        bee_header_0: Optional[bytes] = None,
        bee_header_1: Optional[bytes] = None,
        fcb: Optional[bytes] = None,
        hab_container: Optional[bytes] = None,
    ) -> None:
        """Bootable Image constructor for RT1xxx devices.

        :param family: Chip family.
        :param mem_type: Used memory type.
        :param revision: Chip silicon revision.
        :param bee_header_0: BEE encryption header 0
        :param bee_header_1: BEE encryption header 1
        :param fcb: FCB block, defaults to None
        :param hab_container: Boot image container
        """
        super().__init__(family, mem_type, revision)
        self.bee_header_0 = bee_header_0
        self.bee_header_1 = bee_header_1
        self.fcb = fcb
        if self.fcb:
            self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
            self.fcb_obj.parse(self.fcb)
        self.hab_container = hab_container

    @staticmethod
    def get_segments(mem_type: str) -> List[EnumBimgSegments]:
        """Get list of image segments.

        :param mem_type: Used memory type.
        """
        return [
            EnumBimgSegments.FCB,
            EnumBimgSegments.BEE_HEADER_0,
            EnumBimgSegments.BEE_HEADER_1,
            EnumBimgSegments.HAB_CONTAINER,
        ]

    def parse(self, binary: bytes) -> None:
        """Parse binary into bootable image object.

        :param binary: Full binary of bootable image.
        """
        super().parse(binary)
        if self.fcb:
            self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
            self.fcb_obj.parse(self.fcb)

    def store_config(self, output: str) -> None:
        """Store bootable image into configuration and binary blocks.

        :param output: Path to output folder to store bootable image configuration.
        """
        super().store_config(output)
        if self.fcb_obj:
            write_file(self.fcb_obj.create_config(), os.path.join(output, "fcb.yaml"))

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get list of all supported families by bootable image.

        :return: List of families.
        """
        ignored = BootableImageRt101x.get_supported_families()
        full_list = BootableImage.get_supported_families()
        return [x for x in full_list if re.match(r"[rR][tT]10[\dxX]{2}$", x) and x not in ignored]


class BootableImageRt11xx(BootableImage):
    """Bootable Image class for RT1xxx devices."""

    def __init__(
        self,
        family: str,
        mem_type: str,
        revision: str = "latest",
        keyblob: Optional[bytes] = None,
        fcb: Optional[bytes] = None,
        keystore: Optional[bytes] = None,
        hab_container: Optional[bytes] = None,
    ) -> None:
        """Bootable Image constructor for RT1xxx devices.

        :param family: Chip family.
        :param mem_type: Used memory type.
        :param revision: Chip silicon revision.
        :param keyblob: Key Blob block, defaults to None
        :param fcb: FCB block, defaults to None
        :param keystore: Key store block, defaults to None
        :param hab_container: Boot image container
        """
        super().__init__(family, mem_type, revision)
        self.keyblob = keyblob
        self.fcb = fcb
        self.keystore = keystore
        self.fcb_obj = None
        if self.fcb is not None:
            self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
            self.fcb_obj.parse(self.fcb)
        self.hab_container = hab_container

    @staticmethod
    def get_segments(mem_type: str) -> List[EnumBimgSegments]:
        """Get list of image segments.

        :param mem_type: Used memory type.
        """
        segments = [EnumBimgSegments.HAB_CONTAINER]
        if mem_type == "flexspi_nor":
            segments.extend(
                [
                    EnumBimgSegments.KEYBLOB,
                    EnumBimgSegments.FCB,
                    EnumBimgSegments.KEYSTORE,
                ]
            )
        return segments

    def parse(self, binary: bytes) -> None:
        """Parse binary into bootable image object.

        :param binary: Full binary of bootable image.
        """
        super().parse(binary)
        if self.fcb:
            self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
            self.fcb_obj.parse(self.fcb)

    def store_config(self, output: str) -> None:
        """Store bootable image into configuration and binary blocks.

        :param output: Path to output folder to store bootable image configuration.
        """
        super().store_config(output)
        if self.fcb_obj:
            write_file(self.fcb_obj.create_config(), os.path.join(output, "fcb.yaml"))

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get list of all supported families by bootable image.

        :return: List of families.
        """
        full_list = BootableImage.get_supported_families()
        ignored = BootableImageRt118x.get_supported_families()
        # filter out just RT1xxx
        return [x for x in full_list if re.match(r"[rR][tT]11[\dxX]{2}$", x) and x not in ignored]


class BootableImageRt118x(BootableImage):
    """Bootable Image class for RT118x devices."""

    def __init__(
        self,
        family: str = "rt118x",
        mem_type: str = "flexspi_nor",
        revision: str = "latest",
        keyblob: Optional[bytes] = None,
        fcb: Optional[bytes] = None,
        xmcd: Optional[bytes] = None,
        ahab_container: Optional[bytes] = None,
    ) -> None:
        """Bootable Image constructor for Lpc55s3x devices.

        :param mem_type: Used memory type.
        :param fcb: FCB block, defaults to None
        :param image_version: Image version number, defaults to 0
        :param app: Application block, defaults to None
        """
        assert mem_type == "flexspi_nor"
        super().__init__(family, mem_type, revision)
        self.keyblob = keyblob
        self.fcb = fcb
        self.fcb_obj = None
        if self.fcb:
            self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
            self.fcb_obj.parse(self.fcb)
        self.xmcd = xmcd
        self.xmcd_obj = None
        if self.xmcd:
            self.xmcd_obj = XMCD(self.family, self.revision)
            self.xmcd_obj.parse(self.xmcd)
        self.ahab_container = ahab_container

    @staticmethod
    def get_segments(mem_type: str) -> List[EnumBimgSegments]:
        """Get list of image segments.

        :param mem_type: Used memory type.
        """
        return [
            EnumBimgSegments.KEYBLOB,
            EnumBimgSegments.FCB,
            EnumBimgSegments.XMCD,
            EnumBimgSegments.AHAB_CONTAINER,
        ]

    def store_config(self, output: str) -> None:
        """Store bootable image into configuration and binary blocks.

        :param output: Path to output folder to store bootable image configuration.
        """
        super().store_config(output)
        if self.fcb_obj:
            write_file(self.fcb_obj.create_config(), os.path.join(output, "fcb.yaml"))
        if self.xmcd_obj:
            write_file(self.xmcd_obj.create_config(), os.path.join(output, "xmcd.yaml"))

    def parse(self, binary: bytes) -> None:
        """Parse binary into bootable image object.

        :param binary: Full binary of bootable image.
        """
        # KeyBlob
        offset = self.bimg_descr["keyblob_offset"]
        self.keyblob = binary[offset : offset + self.get_segment_size(EnumBimgSegments.KEYBLOB)]
        # FCB
        offset = self.bimg_descr["fcb_offset"]
        self.fcb = binary[offset : offset + self.get_segment_size(EnumBimgSegments.FCB)]
        self.fcb_obj = FCB(self.family, self.mem_type, self.revision)
        self.fcb_obj.parse(self.fcb)
        # XMCD
        offset = self.bimg_descr["xmcd_offset"]
        size = self._get_xmcd_size(binary[offset : offset + XMCDHeader.SIZE])
        if size > 0:
            self.xmcd = binary[offset:size]
            self.xmcd_obj = XMCD(self.family, self.revision)
            self.xmcd_obj.parse(self.xmcd)
        # AHAB container
        offset = self.bimg_descr["ahab_container_offset"]
        self.ahab_container = binary[offset:]

    def _get_xmcd_size(self, header_binary: bytes) -> int:
        try:
            header = XMCDHeader.parse(header_binary)
        except SPSDKParsingError:
            return 0
        mem_type = MemoryType.name(header.interface)
        config_type = ConfigurationBlockType.name(header.block_type)
        registers = XMCD.load_registers(self.family, mem_type, config_type, self.revision)
        return len(registers.image_info())

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get list of all supported families by bootable image.

        :return: List of supported families.
        """
        return ["rt118x"]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Wrapper for all spsdk applications.

Its purpose is to provide easier discoverability.
New users may not be aware of all available apps.
"""
import sys
from typing import Any

import click

from spsdk import SPSDKError
from spsdk import __version__ as spsdk_version
from spsdk.apps.utils.common_cli_options import CommandsTreeGroup

from .blhost import main as blhost_main
from .dk6prog import main as dk6prog_main
from .elftosb import main as elftosb_main
from .ifr import main as ifr_main
from .nxpcertgen import main as nxpcertgen_main
from .nxpcrypto import main as nxpcrypto_main
from .nxpdebugmbox import main as nxpdebugmbox_main
from .nxpdevhsm import main as nxpdevhsm_main
from .nxpdevscan import main as nxpdevscan_main
from .nxpimage import main as nxpimage_main
from .nxpkeygen import main as nxpkeygen_main
from .pfr import main as pfr_main
from .sdphost import main as sdphost_main
from .sdpshost import main as sdpshost_main
from .shadowregs import main as shadowregs_main

try:
    TP = True
    from .tpconfig import main as tpconfig_main
    from .tphost import main as tphost_main
except SPSDKError:
    TP = False
from .utils.utils import catch_spsdk_error


@click.group(name="spsdk", no_args_is_help=True, cls=CommandsTreeGroup)
@click.version_option(spsdk_version, "--version")
def main() -> int:
    """Main entry point for all SPSDK applications."""
    return 0


main.add_command(blhost_main, name="blhost")
main.add_command(elftosb_main, name="elftosb")
main.add_command(ifr_main, name="ifr")
main.add_command(nxpcertgen_main, name="nxpcertgen")
main.add_command(nxpcrypto_main, name="nxpcrypto")
main.add_command(nxpdebugmbox_main, name="nxpdebugmbox")
main.add_command(nxpdevscan_main, name="nxpdevscan")
main.add_command(nxpdevhsm_main, name="nxpdevhsm")
main.add_command(nxpkeygen_main, name="nxpkeygen")
main.add_command(nxpimage_main, name="nxpimage")
main.add_command(pfr_main, name="pfr")
main.add_command(sdphost_main, name="sdphost")
main.add_command(sdpshost_main, name="sdpshost")
main.add_command(shadowregs_main, name="shadowregs")
main.add_command(dk6prog_main, name="dk6prog")

if TP:
    main.add_command(tpconfig_main, name="tpconfig")
    main.add_command(tphost_main, name="tphost")
else:
    click.echo(
        "Please install SPSDK with pip install 'spsdk[tp]' in order to use tphost and tpconfig apps"
    )


@catch_spsdk_error
def safe_main() -> Any:
    """Call the main function."""
    sys.exit(main())


if __name__ == "__main__":
    safe_main()  # pylint: disable=no-value-for-parameter



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""SerialBuspalProxy serves as patch replacement for serial.Serial class."""

import logging
from typing import (  # pylint: disable=unused-import  # Type is necessary for Mypy
    Dict,
    Optional,
    Type,
)

from spsdk.exceptions import SPSDKError
from spsdk.mboot.interfaces.buspal_i2c import I2cModeCommand
from spsdk.mboot.interfaces.buspal_spi import SpiModeCommand

from .serial_proxy import SerialProxy

logger = logging.getLogger(__name__)


class SerialBuspalProxy(SerialProxy):
    """SerialProxy is used to simulate communication with BUSPAL serial device.

    It can be used as mock.patch for serial.Serial class.
    @patch(<your.package>.Serial, SerialProxy.init_proxy(pre_recorded_responses))
    """

    frame_header: int

    @classmethod
    def init_buspal_proxy(cls, target: str, data: Dict[bytes, bytes]) -> "Type[SerialProxy]":
        """Initialized response dictionary of write and read bytes.

        :param target: BUSPAL target type
        :param data: Dictionary of write and read bytes
        :return: SerialProxy class with configured data
        :raises SPSDKError: target not supported
        """
        if target == "i2c":
            cls.frame_header = I2cModeCommand.write_then_read.value
        elif target == "spi":
            cls.frame_header = SpiModeCommand.write_then_read.value
        else:
            raise SPSDKError(f"Target {target} not supported")
        return super().init_proxy(data)

    def __init__(self, port: str, timeout: int, baudrate: int, write_timeout: Optional[int] = None):
        """Basic initialization for serial.Serial class.

        __init__ signature must accommodate instantiation of serial.Serial

        :param port: Serial port name
        :param timeout: timeout (does nothing)
        :param baudrate: Serial port speed (does nothing)
        """
        self._buffer_index = 0
        self.tx_buffer = b""
        self.rx_buffer = b""
        super().__init__(port, timeout, baudrate, write_timeout=write_timeout)

    def write(self, data: bytes) -> None:
        """Simulates a BUSPAL write, pick up response from responses, store if command frame.

        Description:
            Command frame length is 5 bytes and start with byte 0x8:
                - if byte 1 is not 0, this is the number of bytes to write, sent as consecutive frames
                - if byte 3 is not 0, this is the number of bytes to read, received from consecutive frames

        :param data: Bytes to write, key in responses
        """
        if len(data) == 5 and data[0] == self.frame_header:
            if data[1]:
                self.tx_buffer += data
            elif data[3]:
                self.tx_buffer = b""
        elif self.tx_buffer:
            # append consecutive transmit bytes
            self.tx_buffer += data
            if self.tx_buffer in self.responses:
                self.rx_buffer = self.responses[self.tx_buffer]
                self._buffer_index = 0
        else:
            # not a command frame, store expected response, defined in pre_recorded_responses
            self.rx_buffer = self.responses[data]
            self._buffer_index = 0
        logger.debug(f"[{' '.join(hex(x) for x in data)}]")

    def read(self, length: int) -> bytes:
        """Read portion of pre-configured data.

        :param length: Amount of data to read from buffer
        :return: Data read
        """
        if isinstance(self.rx_buffer, list):
            segment = self.rx_buffer[self._buffer_index][:length]
            self._buffer_index += 1
        else:
            segment = self.rx_buffer[:length]
        logger.debug(f"<{' '.join(hex(x) for x in segment)}>")
        return segment



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for DebugMailbox Pemicro Debug probes support."""

import logging
from typing import Dict, Optional

from pypemicro import PEMicroException, PEMicroInterfaces, PyPemicro

from spsdk.utils.misc import value_to_int

from .debug_probe import (
    DebugProbe,
    SPSDKDebugProbeError,
    SPSDKDebugProbeNotOpenError,
    SPSDKDebugProbeTransferError,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)
PEMICRO_LOGGER = logger.getChild("PyPemicro")


class DebugProbePemicro(DebugProbe):
    """Class to define Pemicro package interface for NXP SPSDK."""

    @staticmethod
    def get_options_help() -> Dict[str, str]:
        """Get full list of options of debug probe.

        :return: Dictionary with individual options. Key is parameter name and value the help text.
        """
        return {
            "frequency": "Set the communication frequency in Hz, default is 100_000Hz",
        }

    @classmethod
    def get_pemicro_lib(cls) -> PyPemicro:
        """Get Pemicro object.

        :return: The Pemicro Object
        :raises SPSDKDebugProbeError: The Pemicro object get function failed.
        """
        try:
            return PyPemicro(
                log_info=PEMICRO_LOGGER.info,
                log_debug=PEMICRO_LOGGER.debug,
                log_err=PEMICRO_LOGGER.error,
                log_war=PEMICRO_LOGGER.warn,
            )
        except PEMicroException as exc:
            raise SPSDKDebugProbeError(f"Cannot get Pemicro library: ({str(exc)})") from exc

    def __init__(self, hardware_id: str, options: Optional[Dict] = None) -> None:
        """The Pemicro class initialization.

        The Pemicro initialization function for SPSDK library to support various DEBUG PROBES.
        """
        super().__init__(hardware_id, options)

        self.pemicro: Optional[PyPemicro] = None

        logger.debug("The SPSDK Pemicro Interface has been initialized")

    @classmethod
    def get_connected_probes(
        cls, hardware_id: Optional[str] = None, options: Optional[Dict] = None
    ) -> list:
        """Get all connected probes over Pemicro.

        This functions returns the list of all connected probes in system by Pemicro package.

        :param hardware_id: None to list all probes, otherwise the the only probe with matching
            hardware id is listed.
        :param options: The options dictionary
        :return: probe_description
        """
        # pylint: disable=import-outside-toplevel
        from .utils import DebugProbes, ProbeDescription

        probes = DebugProbes()
        try:
            connected_probes = PyPemicro.list_ports()
            for probe in connected_probes:
                if not hardware_id or hardware_id == str(probe["id"]):
                    probes.append(
                        ProbeDescription(
                            "PEMicro", probe["id"], probe["description"], DebugProbePemicro
                        )
                    )
        except PEMicroException as exc:
            logger.warning(f"Cannot get list of PEMicro probes: {str(exc)}")
        return probes

    def open(self) -> None:
        """Open Pemicro interface for NXP SPSDK.

        The Pemicro opening function for SPSDK library to support various DEBUG PROBES.
        The function is used to initialize the connection to target and enable using debug probe
        for DAT purposes.

        :raises SPSDKDebugProbeError: The Pemicro cannot establish communication with target
        """
        try:
            self.pemicro = DebugProbePemicro.get_pemicro_lib()
            if self.pemicro is None:
                raise SPSDKDebugProbeError("Getting of Pemicro library failed.")
        except SPSDKDebugProbeError as exc:
            raise SPSDKDebugProbeError(f"Getting of Pemicro library failed({str(exc)}).") from exc
        try:
            self.pemicro.open(debug_hardware_name_ip_or_serialnum=self.hardware_id)
            self.pemicro.connect(PEMicroInterfaces.SWD)
            self.pemicro.set_debug_frequency(value_to_int(self.options.get("frequency", 100000)))
            self.clear_sticky_errors()
            self.power_up_target()

        except PEMicroException as exc:
            raise SPSDKDebugProbeError(
                f"Pemicro cannot establish  communication with target({str(exc)})."
            ) from exc

    def close(self) -> None:
        """Close Pemicro interface.

        The Pemicro closing function for SPSDK library to support various DEBUG PROBES.
        """
        if self.pemicro:
            self.pemicro.close()

    def coresight_reg_read(self, access_port: bool = True, addr: int = 0) -> int:
        """Read coresight register over Pemicro interface.

        The Pemicro read coresight register function for SPSDK library to support various DEBUG PROBES.

        :param access_port: if True, the Access Port (AP) register will be read(default), otherwise the Debug Port
        :param addr: the register address
        :return: The read value of addressed register (4 bytes)
        :raises SPSDKDebugProbeTransferError: The IO operation failed
        :raises SPSDKDebugProbeNotOpenError: The Pemicro probe is NOT opened
        """
        if self.pemicro is None:
            raise SPSDKDebugProbeNotOpenError("The Pemicro debug probe is not opened yet")

        try:
            if access_port:
                ap_ix = (addr & self.APSEL_APBANKSEL) >> self.APSEL_SHIFT
                ret = self.pemicro.read_ap_register(apselect=ap_ix, addr=addr)
            else:
                ret = self.pemicro.read_dp_register(addr=addr)
            return ret
        except PEMicroException as exc:
            self._reinit_target()
            raise SPSDKDebugProbeTransferError(
                f"The Coresight read operation failed({str(exc)})."
            ) from exc

    def coresight_reg_write(self, access_port: bool = True, addr: int = 0, data: int = 0) -> None:
        """Write coresight register over Pemicro interface.

        The Pemicro write coresight register function for SPSDK library to support various DEBUG PROBES.

        :param access_port: if True, the Access Port (AP) register will be write(default), otherwise the Debug Port
        :param addr: the register address
        :param data: the data to be written into register
        :raises SPSDKDebugProbeTransferError: The IO operation failed
        :raises SPSDKDebugProbeNotOpenError: The Pemicro probe is NOT opened
        """
        if self.pemicro is None:
            raise SPSDKDebugProbeNotOpenError("The Pemicro debug probe is not opened yet")

        try:
            if access_port:
                ap_ix = (addr & self.APSEL_APBANKSEL) >> self.APSEL_SHIFT
                self.pemicro.write_ap_register(apselect=ap_ix, addr=addr, value=data)
            else:
                self.pemicro.write_dp_register(addr=addr, value=data)

        except PEMicroException as exc:
            self._reinit_target()
            raise SPSDKDebugProbeTransferError(
                f"The Coresight write operation failed({str(exc)})."
            ) from exc

    def assert_reset_line(self, assert_reset: bool = False) -> None:
        """Control reset line at a target.

        :param assert_reset: If True, the reset line is asserted(pulled down), if False the reset line is not affected.
        :raises SPSDKDebugProbeNotOpenError: The Pemicro debug probe is not opened yet
        :raises SPSDKDebugProbeError: The PyPEMicro probe RESET function failed
        """
        if self.pemicro is None:
            raise SPSDKDebugProbeNotOpenError("The Pemicro debug probe is not opened yet")

        try:
            if assert_reset:
                self.pemicro.control_reset_line(True)
            else:
                self.pemicro.control_reset_line(False)
        except PEMicroException as exc:
            raise SPSDKDebugProbeError(f"Pemicro reset operation failed: {str(exc)}") from exc



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""SB31 module of sbfile."""

from spsdk.sbfile.sb31.commands import (
    CmdCall,
    CmdConfigureMemory,
    CmdCopy,
    CmdErase,
    CmdExecute,
    CmdFillMemory,
    CmdLoad,
    CmdLoadCmac,
    CmdLoadHashLocking,
    CmdLoadKeyBlob,
    CmdProgFuses,
    CmdProgIfr,
    CmdSectionHeader,
    parse_command,
)
from spsdk.sbfile.sb31.constants import EnumCmdTag

__all__ = [
    # commands3
    "CmdErase",
    "CmdLoad",
    "CmdExecute",
    "CmdCall",
    "CmdProgFuses",
    "CmdProgIfr",
    "CmdLoadCmac",
    "CmdLoadHashLocking",
    "CmdCopy",
    "CmdFillMemory",
    "CmdLoadKeyBlob",
    "CmdConfigureMemory",
    "CmdSectionHeader",
    # constants
    "EnumCmdTag",
]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains FCB (Flash Configuration Block) related code."""
import os

from spsdk import SPSDK_DATA_FOLDER

FCB_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "image", "fcb")
FCB_SCH_FILE: str = os.path.join(FCB_DATA_FOLDER, "sch_fcb.yaml")
FCB_DATABASE_FILE: str = os.path.join(FCB_DATA_FOLDER, "database.yaml")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Commands for image module."""

from abc import ABC
from datetime import datetime
from struct import pack, unpack_from
from typing import Any, Iterable, Iterator, List, Mapping, Optional, Tuple, Type, Union

from asn1crypto import cms, util, x509

from spsdk import SPSDKError
from spsdk.crypto import Certificate, Encoding
from spsdk.exceptions import SPSDKValueError
from spsdk.utils.crypto import crypto_backend, matches_key_and_cert
from spsdk.utils.easy_enum import Enum
from spsdk.utils.misc import DebugInfo

########################################################################################################################
# Enums
########################################################################################################################
from .. import SPSDKError
from .header import CmdHeader, CmdTag, Header, SegTag
from .secret import MAC, BaseClass, CertificateImg, EnumAlgorithm, Signature, SrkTable


class EnumWriteOps(Enum):
    """Enum definition for 'flags' control flags in 'par' parameter of Write Data command."""

    WRITE_VALUE = (0, "Write value")
    WRITE_CLEAR_BITS = (1, "Write clear bits")
    CLEAR_BITMASK = (2, "Clear bitmask")
    SET_BITMASK = (3, "Set bitmask")


class EnumCheckOps(Enum):
    """Enum definition for 'par' parameter of Check Data command."""

    ALL_CLEAR = (0, "All bits clear")
    ALL_SET = (1, "All bits set")
    ANY_CLEAR = (2, "Any bit clear")
    ANY_SET = (3, "Any bit set")


class EnumCertFormat(Enum):
    """Certificate format tags."""

    SRK = (0x03, "SRK certificate format")
    X509 = (0x09, "X.509v3 certificate format")
    CMS = (0xC5, "CMS/PKCS#7 signature format")
    BLOB = (0xBB, "SHW-specific wrapped key format")
    AEAD = (0xA3, "Proprietary AEAD MAC format")


class EnumInsKey(Enum):
    """Flags for Install Key commands."""

    CLR = (0, "No flags set")
    ABS = (1, "Absolute certificate address")
    CSF = (2, "Install CSF key")
    DAT = (4, "Key binds to Data Type")
    CFG = (8, "Key binds to Configuration")
    FID = (16, "Key binds to Fabrication UID")
    MID = (32, "Key binds to Manufacturing ID")
    CID = (64, "Key binds to Caller ID")
    HSH = (128, "Certificate hash present")


class EnumAuthDat(Enum):
    """Flags for Authenticate Data commands."""

    CLR = (0, "No flags set")
    ABS = (1, "Absolute signature address")


class EnumEngine(Enum):
    """Engine plugin tags."""

    ANY = (
        0x00,
        "First compatible engine will be selected (no engine configuration parameters are allowed)",
    )
    SCC = (0x03, "Security controller")
    RTIC = (0x05, "Run-time integrity checker")
    SAHARA = (0x06, "Crypto accelerator")
    CSU = (0x0A, "Central Security Unit")
    SRTC = (0x0C, "Secure clock")
    DCP = (0x1B, "Data Co-Processor")
    CAAM = (0x1D, "Cryptographic Acceleration and Assurance Module")
    SNVS = (0x1E, "Secure Non-Volatile Storage")
    OCOTP = (0x21, "Fuse controller")
    DTCP = (0x22, "DTCP co-processor")
    ROM = (0x36, "Protected ROM area")
    HDCP = (0x24, "HDCP co-processor")
    SW = (0xFF, "Software engine")


class EnumCAAM(Enum):
    """CAAM Engine Configuration."""

    DEFAULT = 0x00
    IN_SWAP8 = 0x01
    IN_SWAP16 = 0x02
    OUT_SWAP8 = 0x08
    OUT_SWAP16 = 0x10
    DSC_SWAP8 = 0x40
    DSC_SWAP16 = 0x80


class EnumItm(Enum):
    """Engine configuration flags of Set command."""

    MID = (0x01, "Manufacturing ID (MID) fuse locations")
    ENG = (0x03, "Preferred engine for a given algorithm")


########################################################################################################################
# Abstract Class
########################################################################################################################


class CmdBase:
    """Base class for all commands."""

    def __init__(self, tag: CmdTag, param: int, length: Optional[int] = None):
        """Constructor.

        :param tag: command tag
        :param param: TODO
        :param length: of the binary command representation, in bytes
        """
        self._header = CmdHeader(tag, param, length)

    @property
    def size(self) -> int:
        """Size of command."""
        return self._header.length

    @property
    def tag(self) -> CmdTag:
        """Command tag."""
        return self._header.tag

    @property
    def needs_cmd_data_reference(self) -> bool:
        """Whether the command needs a reference to an additional data.

        If returns True, the following methods must be implemented:
        - cmd_data_offset
        - cmd_data_reference
        """
        return False  # default implementation

    @property
    def cmd_data_offset(self) -> int:
        """Offset of an additional data (such as certificate, signature, etc) in binary image."""
        return 0

    @cmd_data_offset.setter
    def cmd_data_offset(self, value: int) -> None:  # pylint: disable=no-self-use
        """Setter.

        :param value: offset to set
        :raises SPSDKError: If cmd-data not supported by the command
        """
        raise SPSDKError("cmd-data not supported by the command")

    @property
    def cmd_data_reference(self) -> Optional[BaseClass]:
        """Reference to a command data (such as certificate, signature, etc).

        None if no reference was assigned;
        Value type is command-specific
        """
        return None

    @cmd_data_reference.setter
    def cmd_data_reference(self, value: BaseClass) -> None:  # pylint: disable=no-self-use
        """Setter.

        By default, the command does not support cmd_data_reference
        Note: the method must be implemented in `self.has_cmd_data_reference` returns True

        :param value: to be set
        :raises SPSDKError: If reference not supported by the command
        """
        raise SPSDKError("cmd-data not supported by the command")

    def parse_cmd_data(self, data: bytes, offset: int) -> Any:  # pylint: disable=no-self-use
        """Parse additional command data from binary data.

        :param data: to be parsed
        :param offset: start position in data to parse
        :raises SPSDKError: If cmd_data is not supported by the command
        """
        raise SPSDKError("cmd-data not supported by the command")

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and vars(other) == vars(self)

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def info(self) -> str:
        """Text representation of the command."""
        return (
            f'Command "{CmdTag.desc(self.tag)}"   [Tag={str(self.tag)}, length={str(self.size)}]\n'
        )

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Export to binary form (serialization).

        :param dbg_info: debug information about exported data
        :return: binary representation of the command
        """
        hdr_data = self._header.export()
        dbg_info.append_binary_data("header", hdr_data)
        return hdr_data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdBase":
        """Convert binary representation into command (deserialization from binary data).

        :param data: being parsed
        :param offset: current position to readd from data
        :return: parse command
        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")


########################################################################################################################
# HAB Commands
########################################################################################################################


class CmdWriteData(CmdBase):
    """Write data command."""

    @property
    def num_bytes(self) -> int:
        """Number of bytes being written by the command."""
        return self._header.param & 0x7

    @num_bytes.setter
    def num_bytes(self, value: int) -> None:
        """Setter.

        :param value: number of bytes being written by the command
        :raises SPSDKError: When number of bytes is not 1, 2 nor 4
        """
        if value not in (1, 2, 4):
            raise SPSDKError("number of bytes is not 1, 2 nor 4")
        self._header.param &= ~0x7
        self._header.param |= value

    @property
    def ops(self) -> int:
        """Type of write operation."""
        return (self._header.param >> 3) & 0x3

    @ops.setter
    def ops(self, value: int) -> None:
        if value not in EnumWriteOps:
            raise SPSDKValueError("Value not defined")
        self._header.param &= ~(0x3 << 3)
        self._header.param |= int(value) << 3

    def __init__(
        self,
        numbytes: int = 4,
        ops: int = EnumWriteOps.WRITE_VALUE,
        data: Optional[Iterable[Tuple[int, int]]] = None,
    ) -> None:
        """Initialize Write Data command.

        :param numbytes: number of bytes. Must be value: 1, 2 or 4
        :param ops: type of write operation
        :param data: list of tuples: address and value
        :raises SPSDKError: When incorrect number of bytes
        :raises SPSDKError: When incorrect type of operation
        """
        if numbytes not in (1, 2, 4):
            raise SPSDKError("Incorrect number of bytes")
        if ops not in EnumWriteOps:
            raise SPSDKError("Incorrect type of operation")
        super().__init__(CmdTag.WRT_DAT, ((int(ops) & 0x3) << 3) | (numbytes & 0x7))
        self._data: List[List[int]] = []
        if data is not None:
            assert isinstance(data, (list, tuple))
            for address, value in data:
                self.append(address, value)

    def __repr__(self) -> str:
        return f"CmdWriteData <{EnumWriteOps.name(self.ops)}/{self.num_bytes}, {len(self._data)}>"

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, key: int) -> List[int]:
        return self._data[key]

    def __setitem__(self, key: int, value: List[int]) -> None:
        self._data[key] = value

    def __iter__(self) -> Iterator[List[int]]:
        return self._data.__iter__()

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += f"Write Data Command (Ops: {EnumWriteOps.name(self.ops)}, Bytes: {self.num_bytes})\n"
        for cmd in self._data:
            msg += f"- Address: 0x{cmd[0]:08X}, Value: 0x{cmd[1]:08X}\n"
        msg += "-" * 60 + "\n"
        return msg

    def append(self, address: int, value: int) -> None:
        """Append of Write data command."""
        if address < 0 or address > 0xFFFFFFFF:
            raise SPSDKError("Address out of range")
        if value < 0 or value > 0xFFFFFFFF:
            raise SPSDKError("Value out of range")
        self._data.append([address, value])
        self._header.length += 8

    def pop(self, index: int) -> List[int]:
        """Pop of Write data command."""
        if index < 0 or index >= len(self._data):
            raise SPSDKError("Length of data is incorrect")
        cmd = self._data.pop(index)
        self._header.length -= 8
        return cmd

    def clear(self) -> None:
        """Clear of Write data command."""
        self._data.clear()
        self._header.length = self._header.size

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Export to binary form (serialization).

        :param dbg_info: debug information about exported data
        :return: binary representation of the command
        """
        raw_data = super().export(dbg_info=dbg_info)
        for cmd in self._data:
            raw_data += pack(">LL", cmd[0], cmd[1])
        return raw_data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdWriteData":
        """Convert binary representation into command (deserialization from binary data).

        :param data: being parsed
        :param offset: current position to readd from data
        :return: parse command
        """
        header = CmdHeader.parse(data, offset=offset, required_tag=CmdTag.WRT_DAT)
        obj = cls(header.param & 0x7, (header.param >> 3) & 0x3)
        index = header.size
        while index < header.length:
            (address, value) = unpack_from(">LL", data, offset + index)
            obj.append(address, value)
            index += 8
        return obj


class CmdCheckData(CmdBase):
    """Check data command."""

    @property
    def num_bytes(self) -> int:
        """Number of bytes."""
        return self._header.param & 0x7

    @num_bytes.setter
    def num_bytes(self, value: int) -> None:
        if value not in (1, 2, 4):
            raise SPSDKError("Incorrect number of bytes")
        self._header.param &= ~0x7
        self._header.param |= int(value)

    @property
    def ops(self) -> int:
        """Operation of Check data command."""
        return (self._header.param >> 3) & 0x3

    @ops.setter
    def ops(self, value: int) -> None:
        """Operation of Check data command.

        :raises SPSDKError: If incorrect operation
        """
        if value not in EnumCheckOps:
            raise SPSDKError("Incorrect operation")
        self._header.param &= ~(0x3 << 3)
        self._header.param |= int(value) << 3

    def __init__(
        self,
        numbytes: int = 4,
        ops: int = EnumCheckOps.ALL_SET,
        address: int = 0,
        mask: int = 0,
        count: Optional[int] = None,
    ) -> None:
        """Initialize the check data command.

        :param numbytes: number of bytes
        :param ops: type of  operation
        :param address: list of tuples: address and value
        :param mask: mask value
        :param count: count value
        :raises SPSDKError: If incorrect number of bytes
        :raises SPSDKError: If incorrect operation
        """
        if numbytes not in (1, 2, 4):
            raise SPSDKError("Incorrect number of bytes")
        if ops not in EnumCheckOps:
            raise SPSDKError("Incorrect operation")
        super().__init__(CmdTag.CHK_DAT, ((int(ops) & 0x3) << 3) | (numbytes & 0x7))
        self.address = address
        self.mask = mask
        self.count = count
        # the length of 'address'(4B), 'mask'(4B) and count(0 or 4B)  need to be added into Header.length
        self._header.length += 4 + 4 + (4 if count else 0)

    def __repr__(self) -> str:
        return (
            f"CmdCheckData <{EnumCheckOps[self.ops]}/{self.num_bytes}, "  # type: ignore
            f"ADDR=0x{self.address:X}, MASK=0x{self.mask:X}>"
        )

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += f"Check Data Command (Ops: {EnumCheckOps[self.ops]}, Bytes: {self.num_bytes})\n"  # type: ignore

        msg += f"- Address: 0x{self.address:08X}, Mask: 0x{self.mask:08X}"
        if self.count:
            msg += f", Count: {self.count}"
        msg += "\n"
        msg += "-" * 60 + "\n"
        return msg

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Export to binary form (serialization).

        :param dbg_info: debug information about exported data
        :return: binary representation of the command
        """
        raw_data = super().export(dbg_info=dbg_info)
        raw_data += pack(">LL", self.address, self.mask)
        if self.count is not None:
            raw_data += pack(">L", self.count)
        return raw_data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdCheckData":
        """Convert binary representation into command (deserialization from binary data).

        :param data: being parsed
        :param offset: current position to readd from data
        :return: parse command
        """
        header = CmdHeader.parse(data, offset, CmdTag.CHK_DAT)
        numbytes = header.param & 0x7
        ops = (header.param >> 3) & 0x3
        address, mask = unpack_from(">LL", data, offset + header.size)
        count = None
        if (header.length - header.size) > 8:
            count = unpack_from(">L", data, offset + header.size + 8)[0]
        return cls(numbytes, ops, address, mask, count)


class CmdNop(CmdBase):
    """Nop command."""

    def __init__(self, param: int = 0):
        """Initialize the nop command."""
        super().__init__(CmdTag.NOP, param)

    def __repr__(self) -> str:
        return "CmdNop"

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += "-" * 60 + "\n"
        return msg

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdNop":
        """Convert binary representation into command (deserialization from binary data).

        :param data: being parsed
        :param offset: current position to readd from data
        :return: parse command
        """
        header = CmdHeader.parse(data, offset, CmdTag.NOP)
        if header.length != header.size:
            pass
        return cls(header.param)


class CmdSet(CmdBase):
    """Set command."""

    @property
    def itm(self) -> int:
        """Item of Set command."""
        return self._header.param

    @itm.setter
    def itm(self, value: EnumItm) -> None:
        if value not in EnumItm:
            raise SPSDKError("Incorrect item of set command")
        self._header.param = value

    @property
    def hash_algorithm(self) -> EnumAlgorithm:
        """Type of hash algorithm."""
        return self._hash_alg

    @hash_algorithm.setter
    def hash_algorithm(self, value: EnumAlgorithm) -> None:
        if value not in EnumAlgorithm:
            raise SPSDKError("Incorrect type of algorithm")
        self._hash_alg = value

    @property
    def engine(self) -> EnumEngine:
        """Engine plugin tags."""
        return self._engine

    @engine.setter
    def engine(self, value: EnumEngine) -> None:
        if value not in EnumEngine:
            raise SPSDKError("Incorrect type of engine plugin")
        self._engine = value

    def __init__(
        self,
        itm: EnumItm = EnumItm.ENG,
        hash_alg: EnumAlgorithm = EnumAlgorithm.ANY,
        engine: EnumEngine = EnumEngine.ANY,
        engine_cfg: int = 0,
    ):
        """Initialize the set command."""
        if itm not in EnumItm:
            raise SPSDKError("Incorrect engine configuration flag")
        super().__init__(CmdTag.SET, itm)
        self.hash_algorithm: EnumAlgorithm = hash_alg
        self.engine = engine
        self.engine_cfg = engine_cfg
        self._header.length = CmdHeader.SIZE + 4

    def __repr__(self) -> str:
        return (
            f"CmdSet <{EnumItm.name(self.itm)}, {EnumAlgorithm.name(self.hash_algorithm)},"
            f" {EnumEngine.name(self.engine)}, eng_cfg=0x{self.engine_cfg:X}>"
        )

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += "Set Command ITM : {EnumItm.name(self.itm)}\n"
        msg += (
            f"HASH Algo      : {self.hash_algorithm} ({EnumAlgorithm.desc(self.hash_algorithm)})\n"
        )
        msg += f"Engine         : {self.engine} ({EnumEngine.desc(self.engine)})\n"
        msg += f"Engine Conf    : {hex(self.engine_cfg)})\n"
        msg += "-" * 60 + "\n"
        return msg

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Export to binary form (serialization).

        :param dbg_info: debug information about exported data
        :return: binary representation of the command
        """
        raw_data = super().export(dbg_info=dbg_info)
        raw_data += pack("4B", 0x00, self.hash_algorithm, self.engine, self.engine_cfg)
        dbg_info.append_binary_data("data", raw_data)
        return raw_data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdSet":
        """Convert binary representation into command (deserialization from binary data).

        :param data: being parsed
        :param offset: current position to readd from data
        :return: parse command
        """
        header = CmdHeader.parse(data, offset, CmdTag.SET)
        (_, alg, eng, cfg) = unpack_from("4B", data, offset + CmdHeader.SIZE)
        return CmdSet(
            EnumItm.from_int(header.param),
            EnumAlgorithm.from_int(alg),
            EnumEngine.from_int(eng),
            cfg,
        )


class CmdInitialize(CmdBase):
    """Initialize command."""

    @property
    def engine(self) -> int:
        """Engine."""
        return self._header.param

    @engine.setter
    def engine(self, value: EnumEngine) -> None:
        if value not in EnumEngine:
            raise SPSDKError("Incorrect value of engine")
        self._header.param = value

    def __init__(self, engine: int = EnumEngine.ANY, data: Optional[List[int]] = None) -> None:
        """Initialize the initialize command."""
        if engine not in EnumEngine:
            raise SPSDKError("Incorrect value of engine")
        super().__init__(CmdTag.INIT, engine)
        self._data = data if data else []

    def __repr__(self) -> str:
        return f"CmdInitialize <{EnumEngine[self.engine]}, {len(self._data)}>"  # type: ignore

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, key: int) -> int:
        return self._data[key]

    def __setitem__(self, key: int, value: int) -> None:
        self._data[key] = value

    def __iter__(self) -> Iterator[int]:
        return self._data.__iter__()

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += f"Initialize Command (Engine: {EnumEngine[self.engine]})\n"  # type: ignore
        cnt = 0
        for val in self._data:
            msg += f" {cnt:02d}) Value: 0x{val:08X}\n"
            cnt += 1
        msg += "-" * 60 + "\n"
        return msg

    def append(self, value: int) -> None:
        """Appending of Initialize command.

        :raises SPSDKError: If value out of range
        """
        assert isinstance(value, int), "value must be INT type"
        if value < 0 or value >= 0xFFFFFFFF:
            raise SPSDKError("Value out of range")
        self._data.append(value)
        self._header.length += 4

    def pop(self, index: int) -> int:
        """Pop of Initialize command.

        :return: value from the index
        :raises SPSDKError: If incorrect length of data
        """
        if index < 0 or index >= len(self._data):
            raise SPSDKError("Incorrect length of data")
        val = self._data.pop(index)
        self._header.length -= 4
        return val

    def clear(self) -> None:
        """Clear of Initialize command."""
        self._data.clear()
        self._header.length = self._header.size

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Export to binary form (serialization).

        :param dbg_info: debug information about exported data
        :return: binary representation of the command
        """
        raw_data = super().export(dbg_info=dbg_info)
        for val in self._data:
            raw_data += pack(">L", val)
        return raw_data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdInitialize":
        """Convert binary representation into command (deserialization from binary data).

        :param data: being parsed
        :param offset: current position to readd from data
        :return: parse command
        :raises SPSDKError: If incorrect length of data
        """
        header = CmdHeader.parse(data, offset, CmdTag.INIT)
        obj = cls(EnumEngine.from_int(header.param))
        index = header.size
        while index < header.length:
            if (offset + index) >= len(data):
                raise SPSDKError("Incorrect length of data")
            val = unpack_from(">L", data, offset + index)
            obj.append(val[0])
            index += 4
        return obj


class CmdUnlockAbstract(CmdBase, ABC):
    """Abstract unlock engine command; the command depends on engine type."""

    def __init__(self, engine: EnumEngine = EnumEngine.ANY, features: int = 0, uid: int = 0):
        """Constructor.

        :param engine: to be unlocked
        :param features: engine specific features
        :param uid: Unique ID required by some engine/feature combinations
        """
        super().__init__(CmdTag.UNLK, EnumEngine.from_int(engine), length=8)
        self.features = features
        self.uid = uid
        if self._need_uid:
            self._header.length += 8

    def __iter__(self) -> Iterator[int]:
        return self.__iter__()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <{EnumEngine.desc(self.engine)}, {self.features}, {self.uid}>"

    @property
    def engine(self) -> EnumEngine:
        """Engine to be unlocked.

        The term `engine` denotes a peripheral involved in one or more of the following functions:
        - cryptographic computation
        - security state management
        - security alarm handling
        - access control
        """
        return EnumEngine.from_int(self._header.param)

    def info(self) -> str:
        """Text description of the command."""
        msg = super().info()
        msg += "Unlock Command\n"
        msg += f"Engine : {EnumEngine.desc(self.engine)}\n"
        return msg

    @property
    def _need_uid(self) -> bool:
        """Return True if given Engine and Feature requires UID."""
        return self.need_uid(self.engine, self.features)

    @staticmethod
    def need_uid(engine: EnumEngine, features: int) -> bool:
        """Return True if given Engine and Feature requires UID."""
        overall_condition = False
        ocotp_condition = engine == EnumEngine.OCOTP and bool(features & 0b1101)
        overall_condition |= ocotp_condition
        return overall_condition

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdUnlockAbstract":
        """Convert binary representation into command (deserialization from binary data).

        :param data: being parsed
        :param offset: current position to readd from data
        :return: Unlock command
        """
        header = CmdHeader.parse(data, offset, CmdTag.UNLK)
        features = unpack_from(">L", data, offset + header.size)[0]
        engine = EnumEngine.from_int(header.param)
        uid = 0
        if cls.need_uid(engine, features):
            uid = unpack_from(">Q", data, offset + header.size + 4)[0]

        if engine == EnumEngine.SNVS:
            return CmdUnlockSNVS(features)
        if engine == EnumEngine.CAAM:
            return CmdUnlockCAAM(features)
        if engine == EnumEngine.OCOTP:
            return CmdUnlockOCOTP(features, uid)
        return CmdUnlock(engine, features, uid)

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Export to binary form (serialization).

        :param dbg_info: debug information about exported data
        :return: binary representation of the command
        """
        # assert self.size == CmdHeader.SIZE + 4
        raw_data = super().export(dbg_info=dbg_info)
        data = pack(">L", self.features)
        dbg_info.append_binary_data("features", data)
        raw_data += data
        if self._need_uid:
            data = pack(">Q", self.uid)
            dbg_info.append_binary_data("uid", data)
            raw_data += data
        return raw_data


class CmdUnlockSNVS(CmdUnlockAbstract):
    """Command Unlock Secure Non-Volatile Storage (SNVS) Engine."""

    # mask unlock LP_SWR
    FEATURE_UNLOCK_LP_SWR = 1
    # mask unlock ZMK_WRITE
    FEATURE_UNLOCK_ZMK_WRITE = 2

    def __init__(self, features: int = 0) -> None:
        """Constructor.

        :param features: mask of FEATURE_UNLOCK_* constants
        """
        super().__init__(EnumEngine.SNVS, features)

    @property
    def unlock_lp_swr(self) -> bool:
        """Leave LP SW reset unlocked."""
        return self.features & CmdUnlockSNVS.FEATURE_UNLOCK_LP_SWR != 0

    @property
    def unlock_zmk_write(self) -> bool:
        """Leave Zero is able Master Key write unlocked."""
        return self.features & CmdUnlockSNVS.FEATURE_UNLOCK_ZMK_WRITE != 0

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += f"Unlock LP SWR    : {self.unlock_lp_swr}\n"
        msg += f"Unlock ZMK Write : {self.unlock_zmk_write}\n"
        msg += "-" * 60 + "\n"
        return msg


class CmdUnlockCAAM(CmdUnlockAbstract):
    """Command Unlock for Cryptographic Acceleration and Assurance Module ."""

    # Leave Job Ring and DECO Master IP unlocked
    FEATURE_UNLOCK_MID = 1
    # Leave RNG uninitialized
    FEATURE_UNLOCK_RNG = 2
    # Keep manufacturing protection key in internal memory
    FEATURE_UNLOCK_MFG = 4

    def __init__(self, features: int = 0):
        """Initialize.

        :param features: mask of FEATURE_UNLOCK_x constants, defaults to 0
        """
        super().__init__(EnumEngine.CAAM, features)

    @property
    def unlock_mid(self) -> bool:
        """Leave Job Ring and DECO master ID registers unlocked."""
        return self.features & self.FEATURE_UNLOCK_MID != 0

    @property
    def unlock_rng(self) -> bool:
        """Leave RNG un-instantiated."""
        return self.features & self.FEATURE_UNLOCK_RNG != 0

    @property
    def unlock_mfg(self) -> bool:
        """Leave Zero is able Master Key write unlocked."""
        return self.features & self.FEATURE_UNLOCK_MFG != 0

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += f"MID : {self.unlock_mid}\n"
        msg += f"RNG : {self.unlock_rng}\n"
        msg += f"MFG : {self.unlock_mfg}\n"
        msg += "-" * 60 + "\n"
        return msg


class CmdUnlockOCOTP(CmdUnlockAbstract):
    """Command Unlock for On-Chip One-time programable memory (fuses)."""

    # Leave Field Return activation unlocked.
    FEATURE_UNLOCK_FLD_RTN = 1
    # Leave SRK revocation unlocked.
    FEATURE_UNLOCK_SRK_RVK = 2
    # Leave SCS register unlocked.
    FEATURE_UNLOCK_SCS = 4
    # Unlock JTAG using SCS HAB_JDE bit.
    FEATURE_UNLOCK_JTAG = 8

    def __init__(self, features: int = 0, uid: int = 0):
        """Initialize.

        :param features: mask of FEATURE_UNLOCK_x constants, defaults to 0
        :param uid: Unique ID required by some engine/feature combinations
        """
        super().__init__(EnumEngine.OCOTP, features, uid=uid)

    @property
    def _need_uid(self) -> bool:
        """Return True if given Engine and Feature requires UID."""
        return self.unlock_fld_rtn or self.unlock_csc or self.unlock_jtag

    @property
    def unlock_fld_rtn(self) -> bool:
        """Leave Field Return activation unlocked."""
        return self.features & self.FEATURE_UNLOCK_FLD_RTN != 0

    @property
    def unlock_srk_rvk(self) -> bool:
        """Leave SRK revocation unlocked."""
        return self.features & self.FEATURE_UNLOCK_SRK_RVK != 0

    @property
    def unlock_csc(self) -> bool:
        """Leave SCS register unlocked."""
        return self.features & self.FEATURE_UNLOCK_SCS != 0

    @property
    def unlock_jtag(self) -> bool:
        """Unlock JTAG using SCS HAB_JDE bit."""
        return self.features & self.FEATURE_UNLOCK_JTAG != 0

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += f"FLD_RTN : {self.unlock_fld_rtn}\n"
        msg += f"SRK_RVK : {self.unlock_srk_rvk}\n"
        msg += f"CSC     : {self.unlock_csc}\n"
        msg += f"JTAG    : {self.unlock_jtag}\n"
        if self.uid:
            msg += f"UID : {hex(self.uid)}\n"
        msg += "-" * 60 + "\n"
        return msg


class CmdUnlock(CmdUnlockAbstract):
    """Generic unlock engine command."""

    def __init__(self, engine: EnumEngine = EnumEngine.ANY, features: int = 0, uid: int = 0):
        """Constructor.

        :param engine: to be unlocked
        :param features: mask of features to use by the engine
        :param uid: Unique ID (if needed)
        """
        super().__init__(engine, features, uid=uid)

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += f"Features: {self.features})\n"
        msg += f"UID:      {self.uid})\n"
        msg += "-" * 60 + "\n"
        return msg


class CmdInstallKey(CmdBase):
    """Install key command."""

    def __init__(
        self,
        flags: EnumInsKey = EnumInsKey.CLR,
        cert_fmt: EnumCertFormat = EnumCertFormat.SRK,
        hash_alg: EnumAlgorithm = EnumAlgorithm.ANY,
        src_index: int = 0,
        tgt_index: int = 0,
        location: int = 0,
    ) -> None:
        """Constructor.

        :param flags: from EnumInsKey
        :param cert_fmt: format of the certificate; key authentication protocol
        :param hash_alg: hash algorithm
        :param src_index: source key (verification key, KEK) index
        :param tgt_index: target key index
        :param location: start address of an additional data such as KEY to be installed;
                Typically it is relative to CSF start; Might be absolute for DEK key
        """
        super().__init__(CmdTag.INS_KEY, flags)
        self._cert_fmt: EnumCertFormat = cert_fmt
        self.hash_algorithm: EnumAlgorithm = hash_alg
        self.source_index = src_index
        self.target_index = tgt_index
        self.cmd_data_location = location
        self._header.length = CmdHeader.SIZE + 8
        self._certificate_ref: Optional[Union[CertificateImg, SrkTable]] = None

    @property
    def flags(self) -> EnumInsKey:
        """Flags."""
        return EnumInsKey.from_int(self._header.param)

    @flags.setter
    def flags(self, value: EnumInsKey) -> None:
        """Flags.

        :raises SPSDKError: If incorrect flag"
        """
        if value not in EnumInsKey:
            raise SPSDKError("Incorrect flag")
        self._header.param = value

    @property
    def certificate_format(self) -> EnumCertFormat:
        """Certificate format."""
        return self._cert_fmt

    @certificate_format.setter
    def certificate_format(self, value: EnumCertFormat) -> None:
        """Setter.

        :param value: certificate format
        :raises SPSDKError: If incorrect certificate format
        """
        if value not in EnumCertFormat:
            raise SPSDKError("Incorrect certificate format")
        self._cert_fmt = value

    @property
    def hash_algorithm(self) -> EnumAlgorithm:
        """Hash algorithm."""
        return self._hash_alg

    @hash_algorithm.setter
    def hash_algorithm(self, value: EnumAlgorithm) -> None:
        """Setter.

        :param value: hash algorithm
        :raises SPSDKError: If incorrect hash algorithm
        """
        if value not in EnumAlgorithm:
            raise SPSDKError("Incorrect hash algorithm")
        self._hash_alg = value

    @property
    def source_index(self) -> int:
        """Source key (verification key, KEK) index.

        - For SRK, it is index of the SRK key (0-3)
        - For other keys it is index of previously installed target key, typically 0
        """
        return self._src_index

    @source_index.setter
    def source_index(self, value: int) -> None:
        """Setter.

        :param value: source key (verification key, KEK) index
        :raises SPSDKError: If incorrect keys
        :raises SPSDKError: If incorrect keys
        """
        if self._cert_fmt == EnumCertFormat.SRK:
            # RT10xx supports just 4 SRK keys; this might need update for other devices
            if value not in (
                0,
                1,
                2,
                3,
            ):
                raise SPSDKError("Incorrect keys")
        else:
            if value not in (0, 2, 3, 4, 5):
                raise SPSDKError("Incorrect keys")
        self._src_index = value

    @property
    def target_index(self) -> int:
        """Target key index."""
        return self._tgt_index

    @target_index.setter
    def target_index(self, value: int) -> None:
        """Setter.

        :param value: target key index
        :raises SPSDKError: If incorrect key index
        """
        if value not in (0, 1, 2, 3, 4, 5):
            raise SPSDKError("Incorrect key index")
        self._tgt_index = value

    @property
    def cmd_data_offset(self) -> int:
        """Offset of an additional data (such as certificate, signature, etc) in binary image."""
        return self.cmd_data_location

    @cmd_data_offset.setter
    def cmd_data_offset(self, value: int) -> None:
        """Setter.

        :param value: offset to set
        """
        self.cmd_data_location = value

    @property
    def needs_cmd_data_reference(self) -> bool:
        """Whether the command contains a reference to an additional data."""
        if (
            self.flags == EnumInsKey.ABS
        ):  # reference is an absolute address; instance not assigned; used for DEK key
            if self._certificate_ref is not None:
                raise SPSDKError("Reference is not none")
            return False
        return True

    @property  # type: ignore
    def cmd_data_reference(self) -> Optional[Union[CertificateImg, SrkTable]]:
        """Reference to an additional data (such as certificate, signature, etc).

        None if no reference was assigned;
        Value type is command-specific
        """
        return self._certificate_ref

    @cmd_data_reference.setter
    def cmd_data_reference(self, value: Union[CertificateImg, SrkTable]) -> None:
        """Setter.

        By default, the command does not support cmd_data_reference

        :param value: to be set
        """
        assert isinstance(value, (CertificateImg, SrkTable))
        self._certificate_ref = value

    def parse_cmd_data(self, data: bytes, offset: int) -> Union[CertificateImg, SrkTable, None]:
        """Parse additional command data from binary data.

        :param data: to be parsed
        :param offset: start position in data to parse
        :return: parsed data object; command-specific: certificate or SrkTable to be installed
        """
        if self.certificate_format == EnumCertFormat.SRK:
            result: Union[CertificateImg, SrkTable] = SrkTable.parse(data, offset)
        else:
            result = CertificateImg.parse(data, offset)
        self.cmd_data_reference = result
        return result

    @property
    def certificate_ref(self) -> Union[CertificateImg, SrkTable, None]:
        """Corresponding certificate referenced by key-location."""
        return self._certificate_ref

    @certificate_ref.setter
    def certificate_ref(self, value: Union[CertificateImg, SrkTable]) -> None:
        """Setter.

        :param value: certificate to be installed by the command
        """
        self._certificate_ref = value

    def __repr__(self) -> str:
        return (
            f"CmdInstallKey <{EnumInsKey[self.flags]}, {EnumCertFormat[self.certificate_format]},"  # type: ignore
            f" {EnumAlgorithm[self.hash_algorithm]}, {self.source_index}, "  # type: ignore
            f"{self.target_index}, 0x{self.cmd_data_location:X}>"
        )

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += f" Flag      : {self.flags} ({EnumInsKey.desc(self.flags)})\n"
        msg += f" CertFormat: {self.certificate_format}"
        msg += f"({EnumCertFormat.desc(self.certificate_format)})\n"  # type: ignore
        msg += f" Algorithm : {self.hash_algorithm} ({EnumAlgorithm.desc(self.hash_algorithm)})\n"  # type: ignore
        msg += f" SrcKeyIdx : {self.source_index} (Source key index) \n"
        msg += f" TgtKeyIdx : {self.target_index} (Target key index) \n"
        msg += f" Location  : 0x{self.cmd_data_location:08X} (Start address of certificate(s) to install) \n"
        if self.certificate_ref:
            msg += "[related-certificate]\n"
            msg += self.certificate_ref.info()
        msg += "-" * 60 + "\n"
        return msg

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Export to binary form (serialization).

        :param dbg_info: debug information about exported data
        :return: binary representation of the command
        """
        raw_data = super().export(dbg_info=dbg_info)
        data = pack(
            ">4BL",
            self.certificate_format,
            self.hash_algorithm,
            self.source_index,
            self.target_index,
            self.cmd_data_location,
        )
        raw_data += data
        dbg_info.append_binary_data("data", data)
        return raw_data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> CmdBase:
        """Convert binary representation into command (deserialization from binary data).

        :param data: being parsed
        :param offset: current position to read from data
        :return: parse command
        """
        header = CmdHeader.parse(data, offset, CmdTag.INS_KEY)
        protocol, algorithm, src_index, tgt_index, location = unpack_from(
            ">4BL", data, offset + header.size
        )
        return cls(
            EnumInsKey.from_int(header.param),
            protocol,
            algorithm,
            src_index,
            tgt_index,
            location,
        )


# the type represents referenced command data: either Signature or MAC
SignatureOrMAC = Union[MAC, Signature]


class ExpectedSignatureOrMACError(SPSDKError):
    """CmdAuthData additional data block: expected Signature or MAC object."""


class CmdAuthData(CmdBase):
    """Authenticate data command."""

    @property
    def flags(self) -> int:
        """Flag of Authenticate data command."""
        return self._header.param

    @flags.setter
    def flags(self, value: int) -> None:
        if value not in EnumAuthDat:
            raise SPSDKError("Incorrect flag")
        self._header.param = value

    @property
    def key_index(self) -> int:
        """Key index."""
        return self._key_index

    @key_index.setter
    def key_index(self, value: int) -> None:
        if value not in (0, 1, 2, 3, 4, 5):
            raise SPSDKError("Incorrect key index")
        self._key_index = value

    @property
    def engine(self) -> EnumEngine:
        """Engine."""
        return self._engine

    @engine.setter
    def engine(self, value: EnumEngine) -> None:
        if value not in EnumEngine:
            raise SPSDKError("Incorrect engine")
        self._engine = value

    def __init__(
        self,
        flags: EnumAuthDat = EnumAuthDat.CLR,
        key_index: int = 1,
        sig_format: EnumCertFormat = EnumCertFormat.CMS,
        engine: EnumEngine = EnumEngine.ANY,
        engine_cfg: int = 0,
        location: int = 0,
        certificate: Optional[Certificate] = None,
        private_key_pem_data: Optional[bytes] = None,
    ):
        """Initialize the Authenticate data command."""
        super().__init__(CmdTag.AUT_DAT, flags)
        self.key_index = key_index
        self.sig_format = sig_format
        self.engine = engine
        self.engine_cfg = engine_cfg
        self.location = location
        self.certificate = certificate
        self.private_key_pem_data = private_key_pem_data
        self._header.length = CmdHeader.SIZE + 8
        self._blocks: List[Tuple[int, int]] = []  # list of (start-address, size)
        self._signature: Optional[SignatureOrMAC] = None
        if certificate and private_key_pem_data:
            assert isinstance(certificate, Certificate)
            assert isinstance(private_key_pem_data, bytes)
            if not matches_key_and_cert(private_key_pem_data, certificate):
                raise SPSDKError("Given private key does not match the public certificate")

    @property
    def needs_cmd_data_reference(self) -> bool:
        """Whether the command contains a reference to an additional data."""
        return True

    @property
    def cmd_data_offset(self) -> int:
        """Offset of an additional data (such as signature or MAC, etc) in binary image."""
        return self.location

    @cmd_data_offset.setter
    def cmd_data_offset(self, value: int) -> None:
        """Setter.

        :param value: offset to set
        """
        self.location = value

    @property  # type: ignore
    def cmd_data_reference(self) -> Optional[SignatureOrMAC]:
        """Reference to an additional data (such as certificate, signature, etc).

        -   None if no reference was assigned;
        -   Value type is command-specific
        """
        return self._signature

    @cmd_data_reference.setter
    def cmd_data_reference(self, value: SignatureOrMAC) -> None:
        """Setter.

        By default, the command does not support cmd_data_reference

        :param value: to be set
        :raises ExpectedSignatureOrMACError: if unsupported data object is provided
        """
        if self.sig_format == EnumCertFormat.AEAD:
            assert isinstance(value, MAC)
        elif self.sig_format == EnumCertFormat.CMS:
            assert isinstance(value, Signature)
        else:
            raise ExpectedSignatureOrMACError("Unsupported data object is provided")
        self._signature = value

    def parse_cmd_data(self, data: bytes, offset: int) -> SignatureOrMAC:
        """Parse additional command data from binary data.

        :param data: to be parsed
        :param offset: start position in data to parse
        :return: parsed data object; command-specific: Signature or MAC
        :raises ExpectedSignatureOrMACError: if unsupported data object is provided
        """
        header = Header.parse(data, offset)
        if header.tag == SegTag.MAC:
            self._signature = MAC.parse(data, offset)
            return self._signature
        if header.tag == SegTag.SIG:
            self._signature = Signature.parse(data, offset)
            return self._signature
        raise ExpectedSignatureOrMACError(f"TAG = {header.tag}")

    @property
    def signature(self) -> Optional[SignatureOrMAC]:
        """Signature referenced by `location` attribute."""
        return self._signature

    @signature.setter
    def signature(self, value: SignatureOrMAC) -> None:
        """Setter.

        :param value: signature to be installed by the command
        """
        self.cmd_data_reference = value

    def __repr__(self) -> str:
        return (
            f"CmdAuthData <{EnumAuthDat[self.flags]}, {EnumEngine[self.engine]},"  # type: ignore
            f" {self.engine_cfg}, key:{self.key_index}, 0x{self.location:X}>"
        )

    def __len__(self) -> int:
        return len(self._blocks)

    def __getitem__(self, key: int) -> Tuple[int, int]:
        return self._blocks[key]

    def __setitem__(self, key: int, value: Tuple[int, int]) -> None:
        assert isinstance(value, (list, tuple))
        if len(value) != 2:
            raise SPSDKError("Incorrect length")
        self._blocks[key] = value

    def __iter__(self) -> Iterator[Union[Tuple[Any, ...], List[Any]]]:
        return self._blocks.__iter__()

    def info(self) -> str:
        """Text description of the command."""
        msg = "-" * 60 + "\n"
        msg += super().info()
        msg += f" Flag:        {self.flags} ({EnumAuthDat.desc(self.flags)})\n"
        msg += f" Key index:   {self.key_index}\n"
        msg += f" Engine:      {self.engine} ({EnumEngine.desc(self.engine)})\n"
        msg += f" Engine Conf: {self.engine_cfg}\n"
        msg += f" Location:    0x{self.location:08X} (Start address of authentication data) \n"
        if self.signature:
            msg += "[related signature]\n"
            msg += self.signature.info()
        msg += "-" * 60 + "\n"
        for blk in self._blocks:
            msg += f"- Start: 0x{blk[0]:08X}, Length: {blk[1]} Bytes\n"
        return msg

    def append(self, start_address: int, size: int) -> None:
        """Append of Authenticate data command."""
        self._blocks.append(
            (start_address, size),
        )
        self._header.length += 8

    def pop(self, index: int) -> Tuple[int, int]:
        """Pop of Authenticate data command."""
        if index < 0 or index >= len(self._blocks):
            raise SPSDKError("Incorrect length of blocks")
        value = self._blocks.pop(index)
        self._header.length -= 8
        return value

    def clear(self) -> None:
        """Clear of Authenticate data command."""
        self._blocks.clear()
        self._header.length = self._header.size + 8

    def _cms_signature(self, zulu: datetime, data: bytes) -> bytes:
        """Sign provided data and return CMS signature.

        :param zulu: current UTC time+date
        :param data: to be signed
        :return: CMS signature (binary)
        :raises SPSDKError: If certificate is not present
        :raises SPSDKError: If private key is not present
        :raises SPSDKError: If incorrect time-zone"
        """
        if self.certificate is None:
            raise SPSDKError("Certificate is not present")
        if self.private_key_pem_data is None:
            raise SPSDKError("Private key is not present")

        # signed data (main section)
        signed_data = cms.SignedData()
        signed_data["version"] = "v1"
        signed_data["encap_content_info"] = util.OrderedDict([("content_type", "data")])
        signed_data["digest_algorithms"] = [
            util.OrderedDict([("algorithm", "sha256"), ("parameters", None)])
        ]

        # signer info sub-section
        signer_info = cms.SignerInfo()
        signer_info["version"] = "v1"
        signer_info["digest_algorithm"] = util.OrderedDict(
            [("algorithm", "sha256"), ("parameters", None)]
        )
        signer_info["signature_algorithm"] = util.OrderedDict(
            [("algorithm", "rsassa_pkcs1v15"), ("parameters", b"")]
        )
        # signed identifier: issuer amd serial number
        asn1cert = x509.Certificate.load(self.certificate.public_bytes(Encoding.DER))
        signer_info["sid"] = cms.SignerIdentifier(
            {
                "issuer_and_serial_number": cms.IssuerAndSerialNumber(
                    {"issuer": asn1cert.issuer, "serial_number": asn1cert.serial_number}
                )
            }
        )
        # signed attributes
        signed_attrs = cms.CMSAttributes()
        signed_attrs.append(
            cms.CMSAttribute(
                {
                    "type": "content_type",
                    "values": [cms.ContentType("data")],
                }
            )
        )
        # check time-zone is assigned (expected UTC+0)
        if not zulu.tzinfo:
            raise SPSDKError("Incorrect time-zone")
        signed_attrs.append(
            cms.CMSAttribute(
                {
                    "type": "signing_time",
                    "values": [cms.Time(name="utc_time", value=zulu.strftime("%y%m%d%H%M%SZ"))],
                }
            )
        )
        signed_attrs.append(
            cms.CMSAttribute(
                {
                    "type": "message_digest",
                    "values": [cms.OctetString(crypto_backend().hash(data))],  # digest
                }
            )
        )
        signer_info["signed_attrs"] = signed_attrs

        # create signature
        signer_info["signature"] = crypto_backend().rsa_sign(
            self.private_key_pem_data, signed_attrs.dump()
        )

        # Adding SignerInfo object to SignedData object
        signed_data["signer_infos"] = [signer_info]

        # content info
        content_info = cms.ContentInfo()
        content_info["content_type"] = "signed_data"
        content_info["content"] = signed_data

        return content_info.dump()

    def update_signature(
        self, zulu: datetime, data: bytes, base_data_addr: int = 0xFFFFFFFF
    ) -> bool:
        """Update signature.

        This method must be called from parent to provide data to be signed

        :param zulu: current UTC time+date
        :param data: currently generated binary data
        :param base_data_addr: base address of the generated data
        :raises ValueError: When certificate or private key are not assigned
        :raises ValueError: When signatures not assigned explicitly
        :raises SPSDKError: If incorrect start address
        :raises SPSDKError: If incorrect end address
        :raises SPSDKError: If incorrect length
        :return: True if length of the signature was unchanged, as this may affect content of the CSF section (pointer
                        to data);
        """
        if not self.certificate or not self.private_key_pem_data:
            raise SPSDKError("certificate or private key not assigned, cannot update signature")

        if self.signature is None:
            raise SPSDKError(
                "signature must be assigned explicitly, so its version matches to CST version"
            )

        if self._blocks:
            sign_data = b""
            if data:  # if not data specified, create "fake" signature to update length
                total_len = 0
                for blk in self._blocks:
                    start = blk[0] - base_data_addr
                    end = blk[0] + blk[1] - base_data_addr
                    if start < 0:
                        raise SPSDKError("Incorrect start address")
                    if end > len(data):
                        raise SPSDKError("Incorrect end address")
                    sign_data += data[start:end]
                    total_len += blk[1]
                if len(sign_data) != total_len:
                    raise SPSDKError("Incorrect length")
        else:
            sign_data = data  # if no blocks defined, sign complete data; used for CSF
        if isinstance(self.signature, Signature):
            new_signature = self._cms_signature(zulu, sign_data)
            result = len(self.signature.data) == len(new_signature)
            self.signature.data = new_signature
        else:
            assert isinstance(self.signature, MAC)
            result = True
        return result

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Export to binary form (serialization).

        :param dbg_info: debug information about exported data
        :return: binary representation of the command
        """
        self._header.length = self.size
        raw_data = super().export(dbg_info=dbg_info)
        data = pack(
            ">4BL",
            self.key_index,
            self.sig_format,
            self.engine,
            self.engine_cfg,
            self.location,
        )
        dbg_info.append_binary_data("data", data)
        raw_data += data
        for blk in self._blocks:
            blk_data = pack(">2L", blk[0], blk[1])
            dbg_info.append_binary_data("block", blk_data)
            raw_data += blk_data
        return raw_data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdAuthData":
        """Convert binary representation into command (deserialization from binary data).

        :param data: being parsed
        :param offset: current position to read from data
        :return: parse command
        """
        header = CmdHeader.parse(data, offset, CmdTag.AUT_DAT)
        key, sig_format, eng, cfg, location = unpack_from(">4BL", data, offset + header.size)
        obj = cls(
            EnumAuthDat.from_int(header.param),
            key,
            sig_format,
            EnumEngine.from_int(eng),
            cfg,
            location,
        )
        index = header.size + 8
        while index < header.length:
            start_address, size = unpack_from(">2L", data, offset + index)
            obj.append(start_address, size)
            index += 8
        return obj


# mapping of supported commands to the corresponding class
_CMD_TO_CLASS: Mapping[CmdTag, Type[CmdBase]] = {
    CmdTag.WRT_DAT: CmdWriteData,
    CmdTag.CHK_DAT: CmdCheckData,
    CmdTag.NOP: CmdNop,
    CmdTag.SET: CmdSet,
    CmdTag.INIT: CmdInitialize,
    CmdTag.UNLK: CmdUnlockAbstract,
    CmdTag.INS_KEY: CmdInstallKey,
    CmdTag.AUT_DAT: CmdAuthData,
}


def parse_command(data: bytes, offset: int = 0) -> CmdBase:
    """Parse CSF/DCD command.

    :param data: binary data to be parsed
    :param offset: to start parsing
    :return: instance of the command
    :raises SPSDKError: If the command is not valid
    """
    try:
        cmd_tag = CmdTag.from_int(data[offset])
    except ValueError as exc:
        raise SPSDKError("Unknown command at position: " + hex(offset)) from exc
    cmd_class = _CMD_TO_CLASS[cmd_tag]
    return cmd_class.parse(data, offset)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for generating, processing and verifying TP Audit Log."""
import contextlib
import os
import sqlite3
from typing import Iterator, List, NamedTuple, Optional, Tuple, Union

from spsdk.utils.crypto.backend_openssl import ec, openssl_backend

from ..exceptions import SPSDKTpError
from .data_container import Container
from .payload_types import PayloadType

DB_VERSION = 2

CREATE_TABLE_COMMAND = """
    CREATE TABLE IF NOT EXISTS records (
        id integer PRIMARY KEY,
        nxp_cert blob,
        oem_cert_0 blob,
        oem_cert_1 blob,
        oem_cert_2 blob,
        oem_cert_3 blob,
        prod_counter blob,
        start_hash blob,
        signature blob
    );
    CREATE TABLE IF NOT EXISTS properties (
        id integer PRIMARY KEY,
        version integer,
        tp_device_id text
    );
"""

INSERT_COMMAND = """
    INSERT INTO records (
        nxp_cert, oem_cert_0, oem_cert_1, oem_cert_2, oem_cert_3,
        prod_counter, start_hash, signature
    ) VALUES (?, ?, ?, ?, ?, ?, ?, ?)
"""

SELECT_COMMAND = """
    SELECT * from records
"""

# Python is indexing from 0, while SQLite from 1
# Python slice does't include the last item, unlike SQLite's BETWEEN operator
SELECT_SLICE_COMMAND = """
    SELECT * FROM records WHERE id >= (? + 1) and id < (? + 1)
"""

RECORD_COUNT_COMMAND = """
    SELECT COUNT(*) FROM records
"""

INSERT_PROPERTIES_COMMAND = """
    INSERT INTO properties (
        version,  tp_device_id
    ) VALUES (?, ?)
"""

UPDATE_PROPERTIES_COMMAND = """
    UPDATE properties
    SET tp_device_id = ?
"""

SELECT_PROPERTIES_COMMAND = """
    SELECT * from properties
"""


@contextlib.contextmanager
def sqlite_cursor(file_path: str) -> Iterator[sqlite3.Cursor]:
    """Yield an cursor to SQLite database.

    :param file_path: Path to SQLite database file
    :raises SPSDKTpError: Error during SQL operation
    :yield: SQLite cursor
    """
    try:
        conn = sqlite3.connect(file_path)
        cursor = conn.cursor()
        yield cursor
    except sqlite3.Error as sql_error:
        raise SPSDKTpError(
            f"Error during sqlite operation using audit log file '{file_path}': {sql_error}"
        ) from sql_error
    except Exception as e:
        raise SPSDKTpError(str(e)) from e
    finally:
        if "conn" in locals():
            conn.commit()
            conn.close()


def _sqlite_setup(file_path: str, tp_device_id: str) -> None:
    """Setup an SQLite database file."""
    with sqlite_cursor(file_path) as cursor:
        cursor.executescript(CREATE_TABLE_COMMAND)
        cursor.execute(INSERT_PROPERTIES_COMMAND, (DB_VERSION, tp_device_id))


class AuditLogRecord(NamedTuple):
    """Single record in the Audit log."""

    nxp_id_cert: bytes
    oem_id_certs: List[bytes]
    prod_counter: bytes
    start_hash: bytes
    signature: bytes

    @classmethod
    def from_data(cls, container_data: bytes) -> "AuditLogRecord":
        """Create AuditLogRecord from TP Data Container."""
        container = Container.parse(container_data)
        nxp_id_cert = container.get_entry(PayloadType.NXP_DIE_ID_AUTH_CERT).payload
        oem_id_certs = [
            item.payload for item in container.get_entries(PayloadType.OEM_DIE_DEVATTEST_ID_CERT)
        ]
        prod_counter = container.get_entry(PayloadType.OEM_PROD_COUNTER).payload
        start_hash = container.get_entry(PayloadType.OEM_TP_LOG_HASH).payload
        signature = container.get_entry(PayloadType.OEM_TP_LOG_SIGN).payload

        return AuditLogRecord(
            nxp_id_cert=nxp_id_cert,
            oem_id_certs=oem_id_certs,
            prod_counter=prod_counter,
            start_hash=start_hash,
            signature=signature,
        )

    @classmethod
    def from_dict(cls, data: dict) -> "AuditLogRecord":
        """Create AuditLogRecord from a dictionary."""
        try:
            return AuditLogRecord(
                nxp_id_cert=data[PayloadType.name(PayloadType.NXP_DIE_ID_AUTH_CERT)],
                oem_id_certs=data[PayloadType.name(PayloadType.OEM_DIE_DEVATTEST_ID_CERT)],
                prod_counter=data[PayloadType.name(PayloadType.OEM_PROD_COUNTER)],
                start_hash=data[PayloadType.name(PayloadType.OEM_TP_LOG_HASH)],
                signature=data[PayloadType.name(PayloadType.OEM_TP_LOG_SIGN)],
            )
        except Exception as e:
            raise SPSDKTpError(f"Invalid AuditLog dictionary record: {e}") from e

    @property
    def prod_counter_int(self) -> int:
        """Return production counter as an integer."""
        return int.from_bytes(self.prod_counter, byteorder="big")

    def as_dict(self) -> dict:
        """Return dictionary suitable for writing into log file."""
        record = {
            PayloadType.name(PayloadType.NXP_DIE_ID_AUTH_CERT): self.nxp_id_cert,
            PayloadType.name(PayloadType.OEM_DIE_DEVATTEST_ID_CERT): self.oem_id_certs,
            PayloadType.name(PayloadType.OEM_PROD_COUNTER): self.prod_counter,
            PayloadType.name(PayloadType.OEM_TP_LOG_HASH): self.start_hash,
            PayloadType.name(PayloadType.OEM_TP_LOG_SIGN): self.signature,
        }
        return record

    @classmethod
    def from_tuple(cls, data: tuple) -> "AuditLogRecord":
        """Create AuditLogRecord from a tuple."""
        # tuple of length more than 8 signifies using a row directly from table
        # in such case, ignore the first entry 'id'
        if len(data) > 8:
            data = data[1:]
        return AuditLogRecord(
            nxp_id_cert=data[0],
            oem_id_certs=list(filter(None, [data[1], data[2], data[3], data[4]])),
            prod_counter=data[5],
            start_hash=data[6],
            signature=data[7],
        )

    def as_tuple(self) -> tuple:
        """Return AuditLogRecord as tuple."""
        return (
            self.nxp_id_cert,
            *tuple(self.oem_id_certs[i] if i < len(self.oem_id_certs) else None for i in range(4)),
            self.prod_counter,
            self.start_hash,
            self.signature,
        )

    def new_hash(self) -> bytes:
        """Hash together the relevant pieces of the record."""
        data_to_hash = self.nxp_id_cert
        for oem_cert in self.oem_id_certs:
            data_to_hash += oem_cert or bytes()
        data_to_hash += self.prod_counter
        data_to_hash += self.start_hash
        return openssl_backend.hash(data_to_hash)

    def is_valid(self, key: Union[ec.EllipticCurvePublicKey, bytes]) -> bool:
        """Check if record is valid.

        :param key: PEM-encoded public key or public key object
        :return: True if signature checks out
        """
        new_hash = self.new_hash()
        return openssl_backend.ecc_verify(
            public_key=key,
            signature=self.signature,
            data=new_hash,
        )

    def save(self, file_path: str, tp_device_id: str) -> None:
        """Store record in an sqlite database file."""
        if not os.path.isfile(file_path):
            _sqlite_setup(file_path, tp_device_id)
        with sqlite_cursor(file_path) as cursor:
            cursor.execute(INSERT_COMMAND, self.as_tuple())


class AuditLogProperties(NamedTuple):
    """Properties of the Audit log."""

    version: int
    tp_device_id: str

    @classmethod
    def from_tuple(cls, data: tuple) -> "AuditLogProperties":
        """Create AuditLogProperties from database tuple.

        :raises SPSDKTpError: Unexpected number of data members.
        """
        # first item in data is index; we can skip that
        if len(data) != 3:
            raise SPSDKTpError(f"AuditLogProperties record expects 3 items, got: {len(data)}")
        return AuditLogProperties(data[1], data[2])


class AuditLog(List[AuditLogRecord]):
    """Full Audit log, List of AuditLogRecords."""

    @staticmethod
    def load(file_path: str) -> "AuditLog":
        """Load AuditLog from a sqlite database file."""
        log = AuditLog()
        with sqlite_cursor(file_path) as cursor:
            cursor.execute(SELECT_COMMAND)
            rows = cursor.fetchall()
            for row in rows:
                # pylint: disable=no-member  # Pylint struggles to understand Audit log is a list
                log.append(AuditLogRecord.from_tuple(row[1:]))
        return log

    def save(self, file_path: str, tp_device_id: str) -> None:
        """Store AuditLog into a sqlite database file."""
        if os.path.isfile(file_path):
            os.remove(file_path)
        _sqlite_setup(file_path, tp_device_id)
        with sqlite_cursor(file_path) as cursor:
            for record in self:  # pylint: disable=not-an-iterable  # yes, it is iterable
                cursor.execute(INSERT_COMMAND, record.as_tuple())

    @staticmethod
    def records(
        file_path: str, id_slice: Optional[Tuple[int, int]] = None
    ) -> Iterator[Tuple[int, AuditLogRecord]]:
        """Read records from database file.

        :param file_path: Path to database file
        :param id_slice: Read records with id between (x, y), defaults to None
        :yield: Iterator yielding record ID and record itself
        """
        command = (SELECT_SLICE_COMMAND, id_slice) if id_slice else (SELECT_COMMAND,)
        with sqlite_cursor(file_path) as cursor:
            cursor.execute(*command)  # type: ignore
            for item in cursor:
                yield item[0], AuditLogRecord.from_tuple(item)

    @staticmethod
    def record_count(file_path: str) -> int:
        """Return number of records in database file."""
        with sqlite_cursor(file_path) as cursor:
            cursor.execute(RECORD_COUNT_COMMAND)
            row = cursor.fetchone()
            return row[0]

    @staticmethod
    def properties(file_path: str) -> AuditLogProperties:
        """Return properties of database.

        :param file_path: Path to audit log file.
        :raises SPSDKTpError: Unable to read properties table
        """
        try:
            with sqlite_cursor(file_path) as cursor:
                cursor.execute(SELECT_PROPERTIES_COMMAND)
                row = cursor.fetchone()
                return AuditLogProperties.from_tuple(row)
        except SPSDKTpError as e:
            raise SPSDKTpError(
                "Couldn't read audit log properties. Perhaps you're using an old audit log file. "
                f"\nUnderlying issue: {e}"
            ) from e


class AuditLogCounter:
    """Counter for Audit Log stats (records verified, certificates exported)."""

    def __init__(self, check_count: int = 0, nxp_count: int = 0, oem_count: int = 0) -> None:
        """Initialize counters for Audit log verification."""
        self.check_count = check_count
        self.nxp_count = nxp_count
        self.oem_count = oem_count

    def __add__(self, other: "AuditLogCounter") -> "AuditLogCounter":
        if not isinstance(other, AuditLogCounter):
            return NotImplemented
        return AuditLogCounter(
            check_count=self.check_count + other.check_count,
            nxp_count=self.nxp_count + other.nxp_count,
            oem_count=self.oem_count + other.oem_count,
        )

    def __iadd__(self, other: "AuditLogCounter") -> "AuditLogCounter":
        if not isinstance(other, AuditLogCounter):
            return NotImplemented
        self.check_count += other.check_count
        self.oem_count += other.oem_count
        self.nxp_count += other.nxp_count
        return self

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(f'{name}={value}' for name, value in vars(self).items())})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AuditLogCounter):
            return NotImplemented
        return vars(self) == vars(other)

    def __str__(self) -> str:
        return (
            f"Records verified:  {self.check_count:,}\n"
            f"NXP cert exported: {self.nxp_count:,}\n"
            f"OEM cert exported: {self.oem_count:,}"
        )
